import datetime
import errno
import fcntl
import hashlib
import logging
import os
import os.path
import re
import signal
import time

from tornado.escape import json_decode
from tornado.escape import json_encode
from tornado.escape import to_unicode
from tornado.escape import utf8
from tornado.escape import xhtml_escape
from tornado.ioloop import IOLoop

from webtiles import config, connection, game_data_handler, inotify, terminal, util, ws_handler
from webtiles.config import dgl_format_str
from webtiles.connection import WebtilesSocketConnection
from webtiles.game_data_handler import GameDataHandler
from webtiles.inotify import DirectoryWatcher
from webtiles.terminal import TerminalRecorder
from webtiles.util import DynamicTemplateLoader, parse_where_data, PeriodicCallback
from webtiles.ws_handler import CrawlWebSocket, remove_in_lobbys, update_all_lobbys

try:
    from typing import Dict, Set, Tuple, Any
except:
    pass

last_game_id = 0

processes = dict() # type: Dict[str,CrawlProcessHandler]
unowned_process_logger = logging.LoggerAdapter(logging.getLogger(), {})

# this will not template username...
def find_game_info(socket_dir, socket_file):
    game_id = socket_file[socket_file.index(":")+1:-5]
    if (game_id in config.games and
                    os.path.abspath(config.game_param(game_id, "socket_path")) == os.path.abspath(socket_dir)):
        return config.games[game_id]

    game_info = None
    for game_id in config.games:
        if os.path.abspath(config.game_param(game_id, "socket_path")) == os.path.abspath(socket_dir):
            game_info = config.games[game_id]
            break
    return game_info

def handle_new_socket(path, event):
    dirname, filename = os.path.split(path)
    if ":" not in filename or not filename.endswith(".sock"): return
    username = filename[:filename.index(":")]
    abspath = os.path.abspath(path)
    if event == DirectoryWatcher.CREATE:
        if abspath in processes: return # Created by us

        # Find a game_info with this socket path
        game_info = find_game_info(dirname, filename)

        # Create process handler
        process = CrawlProcessHandler(game_info, username,
                                      unowned_process_logger)
        processes[abspath] = process

        try:
            process.connect(abspath)
        except ConnectionRefusedError:
            process.logger.error(
                "Crawl process socket connection refused for %s, socketpath '%s'.",
                game_info["id"], abspath, exc_info=True)
            del processes[abspath]
            return

        process.logger.info("Found a %s game for user %s.", game_info["id"], username)

        # Notify lobbys
        if config.get('dgl_mode'):
            update_all_lobbys(process)
    elif event == DirectoryWatcher.DELETE:
        if abspath not in processes: return
        process = processes[abspath]
        if process.process: return # Handled by us, will be removed later
        process.handle_process_end()
        process.logger.info("Game ended.")
        remove_in_lobbys(process)
        del processes[abspath]

def watch_socket_dirs():
    watcher = DirectoryWatcher()
    added_dirs = set()
    for game_id in config.games:
        socket_dir = os.path.abspath(config.game_param(game_id, "socket_path"))
        if socket_dir in added_dirs:
            continue
        watcher.watch(socket_dir, handle_new_socket)

class CrawlProcessHandlerBase(object):
    def __init__(self, game_params, username, logger):
        self.game_params = game_params
        self.username = username
        self.logger = logging.LoggerAdapter(logger, {})
        try:
            self.logger.manager
            self.logger.process = self._process_log_msg
        except AttributeError:
            # This is a workaround for a python 3.5 bug with chained
            # LoggerAdapters, where delegation is not handled properly (e.g.
            # manager isn't set, _log isn't available, etc.). This simple fix
            # only handles two levels of chaining. The general fix is to
            # upgrade to python 3.7.
            # Issue: https://bugs.python.org/issue31457
            self.logger = logging.LoggerAdapter(logger.logger, {})
            self.logger.process = lambda m,k: logger.process(*self._process_log_msg(m, k))

        self.queue_messages = False

        self.process = None
        self.client_path = self.config_path("client_path")
        self.crawl_version = None
        self.where = {}
        self.wheretime = 0
        self.last_milestone = None
        self.kill_timeout = None

        self.blocked = set()
        self.kicked = {} # not persisted across sessions!

        now = datetime.datetime.utcnow()
        self.formatted_time = now.strftime("%Y-%m-%d.%H:%M:%S")
        self.lock_basename = self.formatted_time + ".ttyrec"

        self.end_callback = None
        self._receivers = set()
        self.last_activity_time = time.time()
        self.idle_checker = PeriodicCallback(self.check_idle, 10000)
        self.idle_checker.start()
        self._was_idle = False
        self.last_watcher_join = 0
        self.receiving_direct_milestones = False

        global last_game_id
        self.id = last_game_id + 1
        last_game_id = self.id

    def _process_log_msg(self, msg, kwargs):
        return "P%-5s %s" % (self.id, msg), kwargs

    def format_path(self, path):
        return dgl_format_str(path, self.username, self.game_params)

    def config_path(self, key):
        if key not in self.game_params:
            return None
        base_path = self.game_params.templated(key, username=self.username)
        if key == "socket_path" and config.get('live_debug'):
            # TODO: this is kind of brute-force given that regular paths aren't
            # validated at all...
            debug_path = os.path.join(base_path, 'live-debug')
            if not os.path.isdir(debug_path):
                os.makedirs(debug_path)
            return debug_path
        else:
            return base_path

    def idle_time(self):
        return int(time.time() - self.last_activity_time)

    def is_idle(self):
        return self.idle_time() > 30

    def check_idle(self):
        if self.is_idle() != self._was_idle:
            self._was_idle = self.is_idle()
            if config.get('dgl_mode'):
                update_all_lobbys(self)

    def flush_messages_to_all(self):
        for receiver in self._receivers:
            receiver.flush_messages()

    def _is_full_map_msg(self, msg):
        # heuristic: map bundles can be very large (100k+), so we don't want to
        # decode. But the crawl process is guaranteed to put some key info
        # first. Use some brute force regexes to check it.
        tocheck = msg[0:50]
        return (re.search(r'"msg" *: *"map"', tocheck)
                and re.search(r'"clear" *: *true', tocheck))

    def _is_spectator_only(self, msg):
        tocheck = msg[0:50]
        return re.search(r'"spect_only" *: *true', tocheck)

    def handle_process_message(self, msg, send): # type: (str, bool) -> None
        # special handling for map messages on a new spectator: these can be
        # massive, and the deflate time adds up, so only send it to new
        # spectators. This is all a bit heuristic, and probably could use
        # a new control message instead...
        # TODO: if multiple spectators join at the same time, it's probably
        # possible for this heuristic to fail and send a full map to everyone
        if self._fresh_watchers and self._is_full_map_msg(msg):
            if self._is_spectator_only(msg):
                for w in self._fresh_watchers:
                    w.append_message(msg, send)
            else:
                for receiver in self._receivers:
                    receiver.append_message(msg, send)
            self._fresh_watchers = set()
            return
        for receiver in self._receivers:
            receiver.append_message(msg, send)

    def send_to_all(self, msg, **data): # type: (str, Any) -> None
        for receiver in self._receivers:
            receiver.send_message(msg, **data)

    def chat_help_message(self, source, command, desc):
        if len(command) == 0:
            self.handle_notification_raw(source,
                        "&nbsp;" * 8 + "<span>%s</span>" % (xhtml_escape(desc)))
        else:
            self.handle_notification_raw(source,
                        "&nbsp;" * 4 + "<span>%s: %s</span>" %
                                (xhtml_escape(command), xhtml_escape(desc)))

    def chat_command_help(self, source):
        # TODO: generalize
        # the chat window is basically fixed width, and these are calibrated
        # to not do a linewrap
        self.handle_notification(source, "The following chat commands are available:")
        self.chat_help_message(source, "/help", "show chat command help")
        self.chat_help_message(source, "/hide", "hide the chat window")
        if self.is_player(source):
            self.chat_help_message(source, "/kick <name> [<minutes>]", "kick <name> for some time")
            self.chat_help_message(source, "", "default: 15 minutes. Not saved across sessions.")
            self.chat_help_message(source, "/block <name>", "kick and block <name>")
            self.chat_help_message(source, "", "Must be present in channel. Special names:")
            self.chat_help_message(source, "", "[anon] disables anonymous spectating")
            self.chat_help_message(source, "", "[all] disables spectating")
            self.chat_help_message(source, "/blocklist", "show your entire blocklist")
            self.chat_help_message(source, "/unblock <name>", "remove <name> from blocklist")
            self.chat_help_message(source, "/unblock *", "clear your blocklist")

    def handle_chat_command(self, source_ws, text):
        # type: (CrawlWebSocket, str) -> bool
        source = source_ws.username
        if len(text) >= 500:
            # sanity check, distinct from chat max length
            text = text[:500]
        text = text.strip()
        if len(text) == 0 or text[0] != '/':
            return False
        splitlist = text.split(None, 1)
        if len(splitlist) == 1:
            command = splitlist[0]
            remainder = ""
        else:
            command, remainder = splitlist
        command = command.lower()
        # TODO: generalize
        if command == "/block" or command == "/mute":
            self.block(source, remainder)
        elif command == "/unblock" or command == "/unmute":
            self.unblock(source, remainder)
        elif command == "/blocklist" or command == "/mutelist":
            self.show_block_list(source)
        elif command == "/help":
            self.chat_command_help(source)
        elif command == "/hide":
            self.hide_chat(source_ws, remainder.strip())
        elif command == "/kick":
            self.kick(source, remainder)
        else:
            return False
        return True

    def is_blocked(self, username):
        if "[all]" in self.blocked:
            return username != self.username
        if not username:
            return "[anon]" in self.blocked
        if username in self.kicked:
            start, interval = self.kicked[username]
            if time.time() - start < interval:
                return True
            else:
                # this doesn't otherwise get cleaned up, so in principle it's
                # a minor memory leak for a long-running process
                del self.kicked[username]
        return username in self.blocked

    def handle_chat_message(self, username, text): # type: (str, str) -> None
        chat_msg = ("<span class='chat_sender'>%s</span>: <span class='chat_msg'>%s</span>" %
                    (username, xhtml_escape(text)))
        self.send_to_all("chat", content = chat_msg)

    def get_receivers_by_username(self, username):
        result = list()
        for w in self._receivers:
            if not w.username:
                continue
            if w.username == username:
                result.append(w)
        return result

    def get_primary_receiver(self):
        # TODO: does this work with console? Probably not...
        if self.username is None:
            return None
        receivers = self.get_receivers_by_username(self.username)
        for r in receivers:
            if not r.watched_game:
                return r
        return None

    def update_receiver_permissions(self):
        for w in self._receivers:
            if self.is_blocked(w.username) and not w.is_admin():
                IOLoop.current().add_callback(
                    lambda w=w: w.go_lobby(message="Spectating this player is restricted."))

    def send_to_user(self, username, msg, **data):
        # type: (str, str, Any) -> None
        # a single user may be viewing from multiple receivers
        for receiver in self.get_receivers_by_username(username):
            receiver.send_message(msg, **data)

    # obviously, don't use this for player/spectator-accessible data. But, it
    # is still partially sanitized in chat.js.
    def handle_notification_raw(self, username, text):
        # type: (str, str) -> None
        msg = ("<span class='chat_msg'>%s</span>" % text)
        self.send_to_user(username, "chat", content=msg, meta=True)

    def handle_notification(self, username, text):
        # type: (str, str) -> None
        self.handle_notification_raw(username, xhtml_escape(text))

    def handle_process_end(self):
        if self.kill_timeout:
            IOLoop.current().remove_timeout(self.kill_timeout)
            self.kill_timeout = None

        self.idle_checker.stop()

        # send game_ended message to watchers. The player is handled in cleanup
        # code in ws_handler.py.
        for watcher in list(self._receivers):
            if watcher.watched_game == self:
                watcher.send_message("game_ended", reason = self.exit_reason,
                                     message = self.exit_message,
                                     dump = self.exit_dump_url)
                # these are individually ok, but with a lot of spectators,
                # doing them in a big loop can easily add up to 100s of ms
                IOLoop.current().add_callback(watcher.go_lobby)

        if self.end_callback:
            self.end_callback()

    def get_watchers(self, chatting_only=False, mark_admins=False):
        # TODO: I don't understand why this code didn't just use self.username,
        # when will this be different than player_name? Maybe for a console
        # player?
        player_name = None
        watchers = list()
        for w in self._receivers:
            if not w.username: # anon
                continue
            if chatting_only and w.chat_hidden:
                continue
            if not w.watched_game:
                player_name = w.username
            else:
                watchers.append(w.username)
                if mark_admins and w.is_admin():
                    watchers[-1] += " (admin)"
        watchers.sort(key=lambda s:s.lower())
        return (player_name, watchers)

    def is_player(self, username):
        # TODO: probably doesn't work for console players spectating themselves
        # TODO: let admin accounts block as well?
        player_name, watchers = self.get_watchers()
        return (username == player_name)

    def hide_chat(self, receiver, param):
        if param == "forever":
            receiver.send_message("super_hide_chat")
            receiver.chat_hidden = True # currently only for super hidden chat
            self.update_watcher_description()
        else:
            receiver.send_message("toggle_chat")

    def restore_blocklist(self, source, l):
        if not self.is_player(source) or l is None:
            return
        if len(l) == 0:
            return
        self.blocked = {u for u in l if u != source}
        self.handle_notification(source, "Restoring blocklist.")
        self.show_block_list(source)
        self.logger.debug("Player '%s' restoring blocklist %s" %
                                            (source, repr(list(self.blocked))))

    def save_blocklist(self, source):
        if not self.is_player(source):
            return
        receiver = self.get_primary_receiver()
        if receiver is not None:
            receiver.save_blocklist(list(self.blocked))

    def kick(self, source, target_args):
        if not self.is_player(source):
            self.handle_notification(source,
                            "You do not have permission to kick spectators.")
            return False

        args = target_args.split(maxsplit=1)
        args.extend([''] * (2 - len(args))) # pad missing args
        target = args[0]
        interval = 15
        try:
            interval = int(args[1])
        except ValueError:
            pass

        if (source == target):
            self.handle_notification(source, "You can't kick yourself!")
            return False
        player_name, watchers = self.get_watchers()
        watchers = set(watchers)

        if target in watchers:
            self.logger.info("Player '%s' has kicked '%s' (%dm)" % (source, target, interval))
            self.handle_notification(source,
                "Spectator '%s' has been kicked for %d minutes." % (target, interval))
        else:
            self.handle_notification(source, "Kick who??")
            return False

        self.kicked[target] = [time.time(), interval * 60]

        receivers = self.get_receivers_by_username(target)
        for w in receivers:
            if not w.is_admin():
                IOLoop.current().add_callback(
                    lambda w=w: w.go_lobby(message="You have been kicked."))
        return True


    def block(self, source, target):
        if not self.is_player(source):
            self.handle_notification(source,
                            "You do not have permission to block spectators.")
            return False
        if (source == target):
            self.handle_notification(source, "You can't block yourself!")
            return False
        player_name, watchers = self.get_watchers()
        watchers = set(watchers)

        if target in watchers:
            self.logger.info("Player '%s' has blocked '%s'" % (source, target))
            self.handle_notification(source,
                                "Spectator '%s' has now been blocked." % target)
        elif target == "[anon]":
            self.logger.info("Player '%s' has disabled anonymous spectating." % source)
            self.handle_notification(source, "Anonymous spectating disabled.")
        elif target == "[all]":
            self.logger.info("Player '%s' has disabled spectating." % source)
            self.handle_notification(source, "Spectating disabled.")
        else:
            # this means that players can't block a username that isn't currently
            # in their chat -- good/bad?
            self.handle_notification(source, "Block who??")
            return False
        self.blocked |= {target}
        self.save_blocklist(source)
        self.update_receiver_permissions() # async
        return True

    def unblock(self, source, target):
        if not self.is_player(source):
            self.handle_notification(source,
                            "You do not have permission to unblock spectators.")
            return False
        if (source == target):
            self.handle_notification(source,
                                    "You can't unblock (or block) yourself!")
            return False
        if target == "*":
            if len(self.kicked):
                self.kicked = {}
                self.handle_notification(source, "Kicks cleared.")
            if (len(self.blocked) == 0):
                self.handle_notification(source, "No one is blocked!")
                return False
            self.logger.info("Player '%s' has cleared their blocklist." % (source))
            self.handle_notification(source, "You have cleared your blocklist.")
            self.blocked = set()
            self.save_blocklist(source)
            self.update_watcher_description()
            return True

        did_something = False
        if target in self.kicked:
            del self.kicked[target]
            self.handle_notification(source, "Removing '%s' from kicks" % target)
            did_something = True

        if not target in self.blocked:
            if not did_something:
                self.handle_notification(source, "Unblock who??")
            return False

        # these messages are more or less ok for [all], [anon]
        self.logger.info("Player '%s' has unblocked '%s'" % (source, target))
        self.handle_notification(source, "You have unblocked '%s'." % target)
        self.blocked -= {target}
        self.save_blocklist(source)
        self.update_watcher_description()
        return True

    def show_block_list(self, source):
        if not self.is_player(source):
            return False
        names = list(self.blocked)
        names.sort(key=lambda s: s.lower())
        if len(names) == 0:
            self.handle_notification(source, "No one is blocked.")
        else:
            self.handle_notification(source, "You have blocked: " +
                                                            ", ".join(names))
        return True

    def get_anon(self):
        return [w for w in self._receivers if not w.username]

    def update_watcher_description(self):
        player_url_template = config.get('player_url')

        def wrap_name(watcher, is_player=False):
            if is_player:
                class_type = 'player'
            else:
                class_type = 'watcher'
            n = watcher
            if player_url_template is None:
                return "<span class='{0}'>{1}</span>".format(class_type, n)
            player_url = player_url_template.replace("%s", watcher.lower())
            username = "<a href='{0}' target='_blank' class='{1}'>{2}</a>".format(player_url, class_type, n)
            return username

        player_name, watchers = self.get_watchers(True, True)

        watcher_names = []
        if player_name is not None:
            watcher_names.append(wrap_name(player_name, True))
        watcher_names += [wrap_name(w) for w in watchers]

        anon_count = len(self.get_anon())
        s = ", ".join(watcher_names)
        if len(watcher_names) > 0 and anon_count > 0:
            s = s + " and %i Anon" % anon_count
        elif anon_count > 0:
            s = "%i Anon" % anon_count
        self.send_to_all("update_spectators",
                         count = self.watcher_count(),
                         names = s)

        if config.get('dgl_mode'):
            update_all_lobbys(self)

    def add_watcher(self, watcher):
        self.last_watcher_join = time.time()
        if self.client_path:
            self._send_client(watcher)
            if watcher.watched_game == self:
                watcher.send_json_options(self.game_params.id, self.username)
        self._receivers.add(watcher)
        self.update_watcher_description()

    def remove_watcher(self, watcher):
        # if both users quit around the same time, this can get out of sync;
        # don't crash in that case
        if watcher in self._receivers:
            self._receivers.remove(watcher)
            self.update_watcher_description()

    def watcher_count(self):
        return len([w for w in self._receivers if w.watched_game and not w.chat_hidden])

    def send_client_to_all(self):
        for receiver in self._receivers:
            self._send_client(receiver)
            if receiver.watched_game == self:
                receiver.send_json_options(self.game_params.id,
                                           self.username)

    def _send_client(self, watcher):
        h = hashlib.sha1(utf8(os.path.abspath(self.client_path)))
        if self.crawl_version:
            h.update(utf8(self.crawl_version))
        v = h.hexdigest()
        GameDataHandler.add_version(v,
                                    os.path.join(self.client_path, "static"))

        templ_path = os.path.join(self.client_path, "templates")
        loader = DynamicTemplateLoader.get(templ_path)
        templ = loader.load("game.html")
        game_html = to_unicode(templ.generate(version = v))
        watcher.send_message("game_client", version = v, content = game_html)

    def stop(self, delay=False):
        if delay:
            IOLoop.current().call_later(0.2, self.stop)
            return
        if self.process:
            self.process.flush_ttyrec()
            try:
                self.process.send_signal(signal.SIGHUP)
            except OSError as e:
                self.logger.error(f"Error {repr(e)} on SIGHUP to child process")
                self._on_process_end()
                return
            t = time.time() + config.get('kill_timeout')
            self.kill_timeout = IOLoop.current().add_timeout(t, self.kill)

    def kill(self):
        if self.process:
            self.logger.info("Killing crawl process after SIGHUP did nothing.")
            try:
                self.process.send_signal(signal.SIGABRT)
            except OSError as e:
                self.logger.error(f"Error {repr(e)} on SIGKILL to child process")
                self._on_process_end()
                return
            finally:
                self.kill_timeout = None

    interesting_info = ("xl", "char", "place", "turn", "dur", "god", "title")

    def set_where_info(self, newwhere):
        # milestone doesn't count as "interesting" but the field is directly
        # handled when sending lobby info by looking at last_milestone
        milestone = bool(newwhere.get("milestone"))
        interesting = (milestone
                or newwhere.get("status") == "chargen"
                or any([self.where.get(key) != newwhere.get(key)
                        for key in CrawlProcessHandlerBase.interesting_info]))

        # ignore milestone sync messages for where purposes
        if newwhere.get("status") != "milestone_only":
            self.where = newwhere
        if milestone:
            self.last_milestone = newwhere
        if interesting:
            update_all_lobbys(self)

    def check_where(self):
        if self.receiving_direct_milestones:
            return
        morgue_path = self.config_path("morgue_path")
        wherefile = os.path.join(morgue_path, self.username + ".where")
        try:
            if os.path.getmtime(wherefile) > self.wheretime:
                self.wheretime = time.time()
                with open(wherefile, "r") as f:
                    wheredata = f.read()

                if wheredata.strip() == "": return

                try:
                    newwhere = parse_where_data(wheredata)
                except:
                    self.logger.warning("Exception while trying to parse where file!",
                                        exc_info=True)
                else:
                    if (newwhere.get("status") == "active" or
                                        newwhere.get("status") == "saved"):
                        self.set_where_info(newwhere)
        except (OSError, IOError):
            pass

    def account_restricted(self):
        # convenience function to check the ws_handler data.
        # this won't work with dgl games, but an account hold sets a dgl ban
        # flag so they can't log in that way
        r = self.get_primary_receiver()
        return r and r.account_restricted()

    def lobby_entry(self):
        u = self.username
        if self.account_restricted():
            u = "[account hold] " + u
        entry = {
            "id": self.id,
            "username": u,
            "spectator_count": self.watcher_count(),
            "idle_time": (self.idle_time() if self.is_idle() else 0),
            "game_id": self.game_params.id,
            }
        for key in CrawlProcessHandlerBase.interesting_info:
            if key in self.where:
                entry[key] = self.where[key]
        if self.last_milestone and self.last_milestone.get("milestone"):
            entry["milestone"] = self.last_milestone.get("milestone")
        return entry

    def human_readable_where(self):
        try:
            return "L{xl} {char}, {place}".format(**self.where)
        except KeyError:
            return ""

    def _base_call(self):
        game = self.game_params
        call = game.get_call_base()

        call += ["-name",   self.username,
                 "-rc",     os.path.join(self.config_path("rcfile_path"),
                                         self.username + ".rc"),
                 "-macro",  os.path.join(self.config_path("macro_path"),
                                         self.username + ".macro"),
                 "-morgue", self.config_path("morgue_path")]

        if self.account_restricted():
            call += ["-no-player-bones"]

        if "options" in game:
            call += game.templated("options", username=self.username)

        if "dir_path" in game:
            call += ["-dir", self.config_path("dir_path")]

        return call

    def note_activity(self):
        self.last_activity_time = time.time()
        self.check_idle()

    def handle_input(self, msg):
        raise NotImplementedError()

class CrawlProcessHandler(CrawlProcessHandlerBase):
    def __init__(self, game_params, username, logger):
        super(CrawlProcessHandler, self).__init__(game_params, username, logger)
        self.socketpath = None
        self.conn = None
        self.ttyrec_filename = None
        self.inprogress_lock = None
        self.inprogress_lock_file = None

        self.exit_reason = None
        self.exit_message = None
        self.exit_dump_url = None

        self._stale_pid = None
        self._stale_lockfile = None
        self._purging_timer = None
        self._process_hup_timeout = None

        self._fresh_watchers = set()

    def start(self):
        self._purge_locks_and_start(True)

    def stop(self, delay=False):
        # n.b. the super call here is partly async
        super(CrawlProcessHandler, self).stop(delay=delay)
        self._stop_purging_stale_processes()
        self._stale_pid = None

    def _purge_locks_and_start(self, firsttime=False):
        # Purge stale locks
        lockfile = self._find_lock()
        if lockfile:
            try:
                with open(lockfile) as f:
                    pid = f.readline()

                try:
                    pid = int(pid)
                except ValueError:
                    # pidfile is empty or corrupted, can happen if the server
                    # crashed. Just clear it...
                    self.logger.error("Invalid PID from lockfile `%s`, clearing", lockfile)
                    self._stale_lockfile = lockfile
                    self._purge_stale_lock()
                    return

                self._stale_pid = pid
                self._stale_lockfile = lockfile
                if firsttime:
                    if self._kill_stale_process(check_pid_only=True):
                        # if we get here, the pid is real
                        hup_wait = 10
                        self.send_to_all("stale_processes",
                                         timeout=hup_wait,
                                         # is name really correct here?
                                         game=self.game_params.templated("name", username=self.username))
                        to = IOLoop.current().add_timeout(time.time() + hup_wait,
                                                          self._kill_stale_process)
                        self._process_hup_timeout = to
                    # else: _purge_stale_lock recurses to this function, so no
                    # need to do anything more here
                else:
                    self._kill_stale_process()
            except Exception:
                self.logger.error("Error while handling lockfile %s.", lockfile,
                                  exc_info=True)
                errmsg = ("Error while trying to terminate a stale process.\n" +
                          "Please contact an administrator.")
                self.exit_reason = "error"
                self.exit_message = errmsg
                self.exit_dump_url = None
                self.handle_process_end()
        else:
            # No more locks, can start
            self._start_process()

    def _stop_purging_stale_processes(self):
        if not self._process_hup_timeout: return
        IOLoop.current().remove_timeout(self._process_hup_timeout)
        self._stale_pid = None
        self._stale_lockfile = None
        self._purging_timer = None
        self._process_hup_timeout = None
        self.handle_process_end()

    def _find_lock(self):
        for path in os.listdir(self.config_path("inprogress_path")):
            if (path.startswith(self.username + ":") and
                path.endswith(".ttyrec")):
                return os.path.join(self.config_path("inprogress_path"),
                                    path)
        return None

    def _kill_stale_process(self, sig=signal.SIGHUP, check_pid_only=False):
        if check_pid_only:
            sig = 0
        self._process_hup_timeout = None
        if self._stale_pid == None:
            return
        if sig == signal.SIGHUP:
            self.logger.info("Purging stale lock at %s, pid %s.",
                             self._stale_lockfile, self._stale_pid)
        elif sig == signal.SIGABRT:
            self.logger.warning("Terminating pid %s forcefully!",
                                self._stale_pid)
        # intentional missing `else`, don't message on 0

        try:
            os.kill(self._stale_pid, sig)
        except ProcessLookupError as e:
            # Process doesn't exist
            self._purge_stale_lock()
            if check_pid_only:
                return False
        except PermissionError as e:
            # Process does exist, but we don't have permissions. Unlikely
            # coincidence between an unrelated process + a stale lock?
            self.logger.error("Clearing stale(?) lock on permission error for pid %i", self._stale_pid)
            self._purge_stale_lock()
            if check_pid_only:
                return False
        except OSError as e:
            self.logger.error("Error while killing process %s.", self._stale_pid,
                              exc_info=True)
            errmsg = ("Error while trying to terminate a stale process.\n" +
                      "Please contact an administrator.")
            self.exit_reason = "error"
            self.exit_message = errmsg
            self.exit_dump_url = None
            self.handle_process_end()
            return False
        else:
            if check_pid_only:
                return True
            if sig == signal.SIGABRT:
                self._purge_stale_lock()
            else:
                if sig == signal.SIGHUP:
                    self._purging_timer = 10
                else:
                    self._purging_timer -= 1

                if self._purging_timer > 0:
                    IOLoop.current().add_timeout(time.time() + 1,
                                                 self._check_stale_process)
                else:
                    self.logger.warning("Couldn't terminate pid %s gracefully.",
                                        self._stale_pid)
                    self.send_to_all("force_terminate?")
                return
        self.send_to_all("hide_dialog")

    def _check_stale_process(self):
        self._kill_stale_process(0)

    def _do_force_terminate(self, answer):
        if answer:
            self._kill_stale_process(signal.SIGABRT)
        else:
            self.handle_process_end()

    def _purge_stale_lock(self):
        if os.path.exists(self._stale_lockfile):
            os.remove(self._stale_lockfile)

        self._purge_locks_and_start(False)

    def _start_process(self):
        self.socketpath = os.path.join(self.config_path("socket_path"),
                                       self.username + ":" +
                                       self.formatted_time + ".sock")

        try: # Unlink if necessary
            os.unlink(self.socketpath)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

        game = self.game_params

        call = self._base_call() + ["-webtiles-socket", self.socketpath,
                                    "-await-connection"]

        ttyrec_path = self.config_path("ttyrec_path")
        if ttyrec_path and config.get('enable_ttyrecs'):
            self.ttyrec_filename = os.path.join(ttyrec_path, self.lock_basename)

        processes[os.path.abspath(self.socketpath)] = self

        if config.get('dgl_mode'):
            self.logger.info("Starting %s.", game["id"])
        else:
            self.logger.info("Starting game.")

        connected = False

        try:
            self.process = TerminalRecorder(call,
                                            self.logger,
                                            config.get('recording_term_size'),
                                            env_vars = game.templated("env", username=self.username, default={}),
                                            game_cwd = game.templated("cwd", username=self.username, default=None),)
            self.process.end_callback = self._on_process_end
            self.process.output_callback = self._on_process_output
            self.process.activity_callback = self.note_activity
            self.process.error_callback = self._on_process_error

            self.process.start(self.ttyrec_filename, self._ttyrec_id_header())

            self.gen_inprogress_lock()

            self.connect(self.socketpath, True)

            self.logger.debug("Crawl FDs: fd%s, fd%s.",
                             self.process.child_fd,
                             self.process.errpipe_read)

            self.last_activity_time = time.time()

            self.check_where()
        except Exception:
            self.logger.warning("Error while starting the Crawl process!", exc_info=True)
            self.exit_reason = "error"
            self.exit_message = "Error while starting the Crawl process!\nSomething has gone very wrong; please let a server admin know."
            self.exit_dump_url = None

            if self.process and self.process.is_started():
                # n.b. we delay a bit here so that the process has more
                # time to start up, and avoid race conditions. (I couldn't come
                # up with anything more reliable.) Also, currently
                # if the connection fails the crawl process will be in a state
                # where it ignores HUP, and the kill handler is needed.
                self.stop(delay=True)
            else:
                self._on_process_end()

    def connect(self, socketpath, primary = False):
        self.socketpath = socketpath
        self.conn = WebtilesSocketConnection(self.socketpath, self.logger)
        self.conn.message_callback = self._on_socket_message
        self.conn.close_callback = self._on_socket_close
        self.conn.username = self.username
        self.conn.connect(primary)

    def gen_inprogress_lock(self):
        self.inprogress_lock = os.path.join(self.config_path("inprogress_path"),
                                            self.username + ":" + self.lock_basename)
        with util.SlowWarning("gen_inprogress_lock '%s'" % self.inprogress_lock):
            f = open(self.inprogress_lock, "w")
            fcntl.lockf(f.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.inprogress_lock_file = f
            cols, lines = self.process.get_terminal_size()
            f.write("%s\n%s\n%s\n" % (self.process.pid, lines, cols))
            f.flush()

    def remove_inprogress_lock(self):
        if self.inprogress_lock_file is None: return
        with util.SlowWarning("remove_inprogress_lock '%s'" % self.inprogress_lock):
            fcntl.lockf(self.inprogress_lock_file.fileno(), fcntl.LOCK_UN)
            self.inprogress_lock_file.close()
            try:
                os.remove(self.inprogress_lock)
            except OSError:
                # Lock already got deleted
                pass

    def _ttyrec_id_header(self): # type: () -> bytes
        clrscr = b"\033[2J"
        crlf = b"\r\n"
        tstamp = int(time.time())
        ctime = time.ctime()
        return (clrscr + b"\033[1;1H" + crlf +
                 utf8("Player: %s" % self.username) + crlf +
                 # odd that this uses name
                 utf8("Game: %s" % self.game_params.templated("name", username=self.username)) + crlf +
                 utf8("Server: %s" % config.get('server_id')) + crlf +
                 utf8("Filename: %s" % self.lock_basename) + crlf +
                 utf8("Time: (%s) %s" % (tstamp, ctime)) + crlf +
                 clrscr)

    def _on_process_end(self):
        if self.process:
            self.logger.debug("Crawl PID %s terminated.", self.process.pid)
        else:
            self.logger.error("Crawl process failed to start, cleaning up.")

        self.remove_inprogress_lock()

        try:
            del processes[os.path.abspath(self.socketpath)]
        except KeyError:
            self.logger.warning("Process entry already deleted: %s", self.socketpath)

        self.process = None

        self.handle_process_end()

    def _on_socket_close(self):
        self.conn = None
        self.stop()

    def handle_process_end(self):
        if self.conn:
            self.conn.close_callback = None
            self.conn.close()
            self.conn = None

        super(CrawlProcessHandler, self).handle_process_end()


    def add_watcher(self, watcher):
        super(CrawlProcessHandler, self).add_watcher(watcher)

        if self.conn and self.conn.open:
            self._fresh_watchers.add(watcher)
            self.conn.send_message('{"msg":"spectator_joined"}')

    def handle_input(self, msg): # type: (str) -> None
        obj = json_decode(msg)

        if obj["msg"] == "input" and self.process:
            self.last_action_time = time.time()

            data = ""
            for x in obj.get("data", []):
                data += chr(x)

            data += obj.get("text", "")

            self.process.write_input(utf8(data))

        elif obj["msg"] == "force_terminate":
            self._do_force_terminate(obj["answer"])

        elif obj["msg"] == "stop_stale_process_purge":
            self._stop_purging_stale_processes()

        elif self.conn and self.conn.open:
            self.conn.send_message(utf8(msg))

    def handle_chat_message(self, username, text): # type: (str, str) -> None
        super(CrawlProcessHandler, self).handle_chat_message(username, text)

        if self.conn and self.conn.open:
            self.conn.send_message(json_encode({
                        "msg": "note",
                        "content": "%s: %s" % (username, text)
                        }))

    def handle_announcement(self, text):
        if self.conn and self.conn.open:
            self.conn.send_message(json_encode({
                        "msg": "server_announcement",
                        "content": text
                        }))

    def _on_process_output(self, line): # type: (str) -> None
        self.check_where()

        try:
            json_decode(line)
        except ValueError:
            self.logger.warning("Invalid JSON output from Crawl process: %s",
                                line)

        # send messages from wrapper scripts only to the player
        for receiver in self._receivers:
            if not receiver.watched_game:
                receiver.append_message(line, True)

    def _on_process_error(self, line): # type: (str) -> None
        morgue_url = self.game_params.templated("morgue_url", username=self.username)
        # The msg this is parsing can be found in dbg-asrt.cc:do_crash_dump
        # this is run line-by-line, so multi-line errors (the norm) may trigger
        # this call more than once
        if line.startswith("ERROR"):
            # header line, e.g. `ERROR in 'wizard.cc' at line 79: Intentional crash`
            self.exit_reason = "crash"
            if line.rfind(":") != -1:
                self.exit_message = line[line.rfind(":") + 1:].strip()
        elif line.find("crash report: ") >= 0:
            self.exit_reason = "crash"
            if morgue_url:
                match = re.search(r"crash report: (.*)", line)
                if match is not None and match.group(1):
                    self.exit_dump_url = morgue_url
                    self.exit_dump_url += os.path.splitext(os.path.basename(match.group(1)))[0]
        elif line.startswith("We crashed!"):
            # before 0.19-a0-1226-g81ff5c4599 everything was on one line; this
            # line prefix is still present but the match below fails.
            self.exit_reason = "crash"
            if morgue_url:
                match = re.search(r"\(([^)]+)\)", line)
                if match is not None:
                    self.exit_dump_url = morgue_url
                    self.exit_dump_url += os.path.splitext(os.path.basename(match.group(1)))[0]
        elif line.startswith("Writing crash info to"):
            # before 0.15-b1-84-gded71f8 the message was very minimal
            self.exit_reason = "crash"
            if morgue_url:
                url = None
                if line.rfind("/") != -1:
                    url = line[line.rfind("/") + 1:].strip()
                elif line.rfind(" ") != -1:
                    url = line[line.rfind(" ") + 1:].strip()
                if url is not None:
                    self.exit_dump_url = morgue_url + os.path.splitext(url)[0]

    def _on_socket_message(self, msg): # type: (str) -> None
        # stdout data is only used for compatibility to wrapper
        # scripts -- so as soon as we receive something on the socket,
        # we stop using stdout
        if self.process:
            self.process.output_callback = None

        if msg.startswith("*"):
            # Special message to the server
            msg = msg[1:]
            msgobj = json_decode(msg)
            if msgobj["msg"] == "client_path":
                if self.client_path == None:
                    # XX Why does this string from the msgobj get templated?
                    self.client_path = self.format_path(msgobj["path"])
                    if "version" in msgobj:
                        self.crawl_version = msgobj["version"]
                        self.logger.info("Crawl version: %s.", self.crawl_version)
                    self.send_client_to_all()
            elif msgobj["msg"] == "flush_messages":
                # only queue, once we know the crawl process asks for flushes
                # note: every version since 0.13 supports this
                self.queue_messages = True;
                self.flush_messages_to_all()
            elif msgobj["msg"] == "dump":
                morgue_url = self.game_params.templated("morgue_url", username=self.username)
                if morgue_url:
                    url = morgue_url + msgobj["filename"]
                    if msgobj["type"] == "command":
                        self.send_to_all("dump", url = url)
                    else:
                        self.exit_dump_url = url
            elif msgobj["msg"] == "exit_reason":
                self.exit_reason = msgobj["type"]
                if "message" in msgobj:
                    self.exit_message = msgobj["message"]
                else:
                    self.exit_message = None
            elif msgobj["msg"] == "milestone":
                # milestone/whereis update: milestone fields are right in the
                # message
                self.receiving_direct_milestones = True # no need for .where files
                self.set_where_info(msgobj)
            else:
                self.logger.warning("Unknown message from the crawl process: %s",
                                    msgobj["msg"])
        else:
            self.check_where()
            if time.time() > self.last_watcher_join + 2:
                # Treat socket messages as activity, since it's otherwise
                # hard to determine activity for games found via
                # watch_socket_dirs.
                # But don't if a spectator just joined, since we don't
                # want that to reset idle time.
                self.note_activity()

            self.handle_process_message(msg, not self.queue_messages)



class DGLLessCrawlProcessHandler(CrawlProcessHandler):
    def __init__(self, logger):
        # used when dgl_mode = False
        game_params = config.GameConfig(dict(
            name = "DCSS",
            id = "dcss-webtiles",
            ttyrec_path = "./",
            inprogress_path = "./",
            socket_path = "./",
            client_path = "./webserver/game_data"))
        super(DGLLessCrawlProcessHandler, self).__init__(game_params,
                                                         "game",
                                                         logger)

    def _base_call(self):
        return ["./crawl"]

    def check_where(self):
        pass



import fcntl
import os
import os.path
import socket
import tempfile
import time
import warnings
from datetime import datetime
from datetime import timedelta

from tornado.escape import json_encode
from tornado.escape import to_unicode
from tornado.escape import utf8
from tornado.ioloop import IOLoop

from webtiles import config, util


class WebtilesSocketConnection(object):
    def __init__(self, socketpath, logger):
        self.crawl_socketpath = socketpath
        self.logger = logger
        self.message_callback = None
        self.socket = None
        self.socketpath = None
        self.open = False
        self.close_callback = None
        self.username = ""

        self.msg_buffer = None

    def connect(self, primary = True):
        if not os.path.exists(self.crawl_socketpath):
            # Wait until the socket exists
            IOLoop.current().add_timeout(time.time() + 1, self.connect)
            return

        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.socket.setblocking(False)

        # Set close-on-exec
        flags = fcntl.fcntl(self.socket.fileno(), fcntl.F_GETFD)
        fcntl.fcntl(self.socket.fileno(), flags | fcntl.FD_CLOEXEC)

        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if (self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF) < 2048):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)
        # on linux, the following may have no effect (setting SO_RCVBUF is
        # often documented as having no effect), but on other unixes, it
        # matters quite a bit. The choice of 212992 is based on the linux
        # default.
        if (self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF) < 212992):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 212992)

        # Bind to a temp path
        # there's a race condition here...
        # note that mktmp here is deprecated, and we may eventually need to
        # do something different. One simple idea is to keep sockets in a
        # temporary directory generated from tempfile calls (i.e
        # tempfile.mkdtemp), but use our own naming scheme. Because this is a
        # socket, regular calls in tempfile are not appropriate.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.socketpath = tempfile.mktemp(dir=config.get('server_socket_path'),
                                              prefix="crawl", suffix=".socket")
        self.socket.bind(self.socketpath)

        # Install handler
        IOLoop.current().add_handler(self.socket.fileno(),
                                     self._handle_read,
                                     IOLoop.ERROR | IOLoop.READ)

        msg = json_encode({
                "msg": "attach",
                "primary": primary
                })

        self.open = True

        self.send_message(utf8(msg))

    def _handle_read(self, fd, events):
        if events & IOLoop.READ:
            data = self.socket.recv(128 * 1024, socket.MSG_DONTWAIT)

            self._handle_data(data)

        if events & IOLoop.ERROR:
            pass

    def _handle_data(self, data): # type: (bytes) -> None
        if self.msg_buffer is not None:
            data = self.msg_buffer + data

        # TODO: is this check safe? Decoding won't always work for
        # fragmented messages...
        if data[-1] != b'\n'[0]:
            # All messages from crawl end with \n.
            # If this one doesn't, it's fragmented.
            self.msg_buffer = data
        else:
            self.msg_buffer = None

            if self.message_callback:
                self.message_callback(to_unicode(data))

    def send_message(self, data): # type: (str) -> None
        start = datetime.now()
        try:
            self.socket.sendto(utf8(data), self.crawl_socketpath)
        except socket.timeout:
            self.logger.warning("Game socket send timeout", exc_info=True)
            self.close()
            return
        except FileNotFoundError:
            # I *think* this may be a relatively normal thing to happen, in
            # which case it probably doesn't need a warning...
            self.logger.warning("Game socket closed during sendto")
            self.close()
            return
        end = datetime.now()
        if end - start >= timedelta(seconds=1):
            self.logger.warning("Slow socket send: " + str(end - start))

    def close(self):
        if self.socket:
            IOLoop.current().remove_handler(self.socket.fileno())
            self.socket.close()
            os.remove(self.socketpath)
            self.socket = None
        if self.close_callback:
            self.close_callback()



#!/usr/bin/env python3

"""
Generate species-data.h, aptitudes.h, species-groups.h, and species-type.h

Works with both Python 2 & 3. If that changes, update how the Makefile calls
this.
"""

from __future__ import print_function

import argparse
import os
import sys
import traceback
import re
import collections
if sys.version_info.major == 2:
    from collections import MutableMapping
else:
    from collections.abc import MutableMapping

import yaml  # pip install pyyaml

from genutil import *

class Species(MutableMapping):
    """Parser for YAML definition files.

    If any YAML content is invalid, the relevant parser function below should
    raise ValueError.
    """

    # TODO: unify with processing in from_yaml
    YAML_MAIN_FIELDS = {'TAG_MAJOR_VERSION', 'fake_mutations', 'difficulty',
            'recommended_jobs', 'enum', 'monster', 'name', 'short_name',
            'adjective', 'genus', 'species_flags', 'aptitudes', 'can_swim',
            'undead_type', 'size', 'str', 'int', 'dex', 'levelup_stats',
            'levelup_stat_frequency', 'recommended_jobs', 'recommended_weapons',
            'difficulty', 'difficulty_priority', 'create_enum', 'walking_verb',
            'altar_action', 'mutations', 'child_name', 'orc_name',
            'orcification_msg'}

    def __init__(self, yaml_dict):
        self.backing_dict = dict()
        self.from_yaml(yaml_dict)

    def __getitem__(self, key):
        return self.backing_dict[key]

    def __setitem__(self, key, val):
        self.backing_dict[key] = val

    def __delitem__(self, key):
        del self.backing_dict[key]

    def __iter__(self):
        return iter(self.backing_dict)

    def __len__(self):
        return len(self.store)

    def set_recommended_weapons(self, weapons):
        if not self.starting_species:
            self.backing_dict['recommended_weapons'] = ""
            return
        if not weapons:
            weapons = list(ALL_WEAPON_SKILLS)
            weapons.remove('short blades')
            weapons.remove('unarmed combat')
        out = []
        for weap in weapons:
            if weap.lower() == "maces and flails":
                weap = "maces flails"
            out.append(enumify(validate_string(weap, 'Weapon Skill',
                                          '[A-Za-z_ ]+'), SKILL_ENUM))
        self.backing_dict['recommended_weapons'] = ", ".join(out)

    def print_unknown_warnings(self, s):
        for key in s:
            if key not in self.YAML_MAIN_FIELDS:
                print("species_gen.py warning: Unknown field '%s' in species %s"
                                    % (key, self['enum']), file=sys.stderr)

    def levelup_stats_from_yaml(self, s):
        self['levelup_stats'] = levelup_stats(s.get('levelup_stats', "default"))
        self['levelup_stat_frequency'] = validate_int_range(
                s['levelup_stat_frequency'], 'levelup_stat_frequency', 0, 28)

        if (self['levelup_stats'] == empty_set("stat_type")
                                    and self['levelup_stat_frequency'] < 28):
            print("species_gen.py warning: species %s has empty levelup_stats"
                  " but a <28 levelup_stat_frequency."
                  % (self['enum']), file=sys.stderr)

        if (self['levelup_stats'] != empty_set("stat_type")
                                    and self['levelup_stat_frequency'] > 27):
            print("species_gen.py warning: species %s has non-empty"
                  " levelup_stats but a levelup_stat_frequency that is > 27."
                  % (self['enum']), file=sys.stderr)

    def from_yaml(self, s):
        # Pre-validation
        if s.get('TAG_MAJOR_VERSION', None) is not None:
            if not isinstance(s['TAG_MAJOR_VERSION'], int):
                raise ValueError('TAG_MAJOR_VERSION must be an integer')
        if not isinstance(s.get('fake_mutations', []), list):
            raise ValueError('fake_mutations must be a list')
        self.starting_species = s.get('difficulty') != False
        has_recommended_jobs = bool(s.get('recommended_jobs'))
        if self.starting_species != has_recommended_jobs:
            raise ValueError('recommended_jobs must not be empty (or'
                                                ' difficulty must be False)')

        # Set attributes
        self['enum'] = validate_string(s['enum'], 'enum', 'SP_[A-Z_]+$')
        self['monster_name'] = validate_string(s['monster'], 'monster',
                                                'MONS_[A-Z_]+$')
        self['name'] = validate_string(s['name'], 'name', '..+')
        self['short_name'] = s.get('short_name', s['name'][:2])
        self['adjective'] = quote_or_nullptr('adjective', s)
        self['genus'] = quote_or_nullptr('genus', s)
        self['species_flags'] = species_flags(s.get('species_flags', []))
        self['xp'] = validate_int_range(s['aptitudes']['xp'], 'xp', -10, 10)
        self['hp'] = validate_int_range(s['aptitudes']['hp'], 'hp', -10, 10)
        self['mp'] = validate_int_range(s['aptitudes']['mp_mod'], 'mp_mod',
                                                                        -5, 20)
        self['wl'] = validate_int_range(s['aptitudes']['wl'], 'wl', 0, 20)
        self['aptitudes'] = aptitudes(s['aptitudes'])
        self['habitat'] = 'HT_LAND' if not s.get('can_swim') else 'HT_WATER'
        self['undead'] = undead_type(s.get('undead_type', 'alive'))
        self['size'] = size(s.get('size', 'medium'))
        self['str'] = validate_int_range(s['str'], 'str', 1, 100)
        self['int'] = validate_int_range(s['int'], 'int', 1, 100)
        self['dex'] = validate_int_range(s['dex'], 'dex', 1, 100)
        self.levelup_stats_from_yaml(s)
        self['mutations'] = mutations(s.get('mutations', {}))
        self['fake_mutations_long'] = fake_mutations_long(
                                            s.get('fake_mutations', []))
        self['fake_mutations_short'] = fake_mutations_short(
                                            s.get('fake_mutations', []))
        self['recommended_jobs'] = recommended_jobs(
                                            s.get('recommended_jobs', []))
        self.set_recommended_weapons(s.get('recommended_weapons', []))
        self['difficulty'] = difficulty(s.get('difficulty'))
        self['difficulty_priority'] = validate_int_range(difficulty_priority(
            s.get('difficulty_priority', 0)), 'difficulty_priority', 0, 1000)
        self['create_enum'] = validate_bool(
                                    s.get('create_enum', False), 'create_enum')
        self['walking_verb'] = quote_or_nullptr('walking_verb', s)
        self['altar_action'] = quote_or_nullptr('altar_action', s)
        self['child_name']   = quote_or_nullptr('child_name', s)
        self['orc_name']     = quote_or_nullptr('orc_name', s)
        self['orcification_msg'] = quote_or_nullptr('orcification_msg', s)

        if 'TAG_MAJOR_VERSION' in s:
            self['tag_major_version_opener'] = (
                        "#if TAG_MAJOR_VERSION == %s" % s['TAG_MAJOR_VERSION'])
            self['tag_major_version_closer'] = "#endif"
        else:
            self['tag_major_version_opener'] = ''
            self['tag_major_version_closer'] = ''
        self.print_unknown_warnings(s)

SpeciesGroup = collections.namedtuple('SpeciesGroup',
                                            ['position', 'width', 'species'])
SpeciesGroupEntry = collections.namedtuple('SpeciesGroupEntry',
                                            ['priority', 'enum'])
SPECIES_GROUPS_TEMPLATE = collections.OrderedDict()
SPECIES_GROUPS_TEMPLATE['Simple'] = SpeciesGroup('coord_def(0, 0)', '50', [])
SPECIES_GROUPS_TEMPLATE['Intermediate'] = SpeciesGroup('coord_def(1, 0)', '20', [])
SPECIES_GROUPS_TEMPLATE['Advanced'] = SpeciesGroup('coord_def(2, 0)', '20', [])
SPECIES_GROUP_TEMPLATE = """
    {{
        "{name}",
        {position},
        {width},
        {{ {species} }}
    }},
"""
ALL_APTITUDES = ('fighting', 'short_blades', 'long_blades', 'axes',
    'maces_and_flails', 'polearms', 'staves', 'ranged weapons',
    'throwing', 'armour', 'dodging', 'stealth', 'shields', 'unarmed_combat',
    'spellcasting', 'conjurations', 'hexes', 'summoning',
    'necromancy', 'forgecraft', 'translocations', 'fire_magic',
    'ice_magic', 'air_magic', 'earth_magic', 'alchemy', 'invocations',
    'evocations', 'shapeshifting')
UNDEAD_TYPES = ('alive', 'undead', 'semi_undead')
SIZES = ('SIZE_TINY', 'SIZE_LITTLE', 'SIZE_SMALL', 'SIZE_MEDIUM', 'SIZE_LARGE',
    'SIZE_GIANT')
ALL_STATS = ('str', 'int', 'dex')
ALL_WEAPON_SKILLS = ('short blades', 'long blades', 'axes',
    'maces and flails', 'polearms', 'staves', 'unarmed combat')

ALL_SPECIES_FLAGS = {'draconian', 'no_hair', 'no_bones',
    'small_torso', 'barding', 'no_feet',
    'no_blood', 'no_ears'}

JOB_ENUM = 'JOB'
UNDEAD_TYPE_ENUM = 'US'
SPECIES_FLAG_ENUM = 'SPF'
SKILL_ENUM = 'SK'

def recommended_jobs(jobs):
    return ', '.join(enumify(validate_string(job, 'Job', '[A-Za-z_ ]+'),
                             JOB_ENUM) for job in jobs)


def size(size):
    val = "SIZE_%s" % size.upper()
    if val not in SIZES:
        raise ValueError('Size %s is invalid, pick one of tiny, little, '
                                    'small, medium, large, big, or giant')
    return val


def species_flags(flags):
    global ALL_SPECIES_FLAGS
    out = set()
    for f in flags:
        if f not in ALL_SPECIES_FLAGS:
            raise ValueError("Unknown species flag %s" % f)
        out.add(enumify(f, SPECIES_FLAG_ENUM))
    if not out:
        out.add('SPF_NONE')
    return ' | '.join(out)


def undead_type(type):
    if type not in UNDEAD_TYPES:
        raise ValueError('Unknown undead type %s' % type)
    return enumify(type, UNDEAD_TYPE_ENUM)


def levelup_stats(stats):
    if stats == "default":
        stats = ALL_STATS
    else:
        # this is pretty loose type checking because we don't want to make
        # any assumptions about how yaml parser handles sequences.
        if isinstance(stats, str):
            raise ValueError(
                "Expected `default` or list for levelup_stats, not `%s`" % stats)
        for s in stats:
            if s not in ALL_STATS:
                raise ValueError('Unknown stat %s' % s)
    if len(stats) == 0:
        return empty_set("stat_type")
    else:
        return make_list(', '.join("STAT_%s" % s.upper() for s in stats))


def mutations(mut_def):
    out = []
    for xl, muts in sorted(mut_def.items()):
        validate_int_range(xl, 'Mutation Level', 1, 27)
        if not isinstance(muts, dict):
            raise ValueError('Mutation key %s doesn\'t seem to have a valid '
                                        'map of {name: amount} entries' % xl)
        for mut_name, amt in sorted(muts.items()):
            validate_string(mut_name, 'Mutation Name', 'MUT_[A-Z_]+')
            validate_int_range(amt, 'Mutation Amount', -3, 3)
            out.append("{{ {mut_name}, {amt}, {xl} }}".format(
                mut_name=mut_name,
                xl=xl,
                amt=amt,
            ))
    return make_list(', '.join(out))

def fake_mutations_long(fmut_def):
    return make_list(', '.join(quote(m.get('long'))
                                    for m in fmut_def if m.get('long')))

def fake_mutations_short(fmut_def):
    return make_list(', '.join(quote(m.get('short'))
                                    for m in fmut_def if m.get('short')))

def aptitudes(apts):
    for apt, val in apts.items():
        if apt not in ALL_APTITUDES and apt not in ('xp', 'hp', 'mp_mod', 'wl'):
            raise ValueError("Unknown aptitude (typo?): %s" % apt)
        validate_int_range(val, apt, -10, 20)
    return apts


def difficulty(d):
    if d not in SPECIES_GROUPS_TEMPLATE.keys() and d is not False:
        raise ValueError("Unknown difficulty: %s" % d)
    return d


def difficulty_priority(prio):
    try:
        return int(prio)
    except ValueError:
        raise ValueError('difficulty_priority value "%s" is not an integer' %
                                prio)


def generate_aptitudes_data(s, template):
    """Convert a species in YAML representation to aptitudes.h format.

    If any of the required data can't be loaded, ValueError is raised and
    passed to the caller.
    """
    # Now generate the aptitudes block. The default is 0.
    # Note: We have to differentiate between 0 and 'False' aptitudes specified
    # in YAML. The latter is UNUSABLE_SKILL.
    aptitudes = {apt: 0 for apt in ALL_APTITUDES}
    for apt, val in s['aptitudes'].items():
        if apt in ('xp', 'hp', 'mp_mod', 'wl'):
            continue
        if val is False:
            aptitudes[apt] = 'UNUSABLE_SKILL'
        else:
            aptitudes[apt] = val
    aptitudes['tag_major_version_opener'] = s['tag_major_version_opener']
    aptitudes['tag_major_version_closer'] = s['tag_major_version_closer']
    return template.format(enum = s['enum'], **aptitudes)


def update_species_group(sg, s):
    difficulty = s['difficulty']
    if difficulty is False:
        # Don't add this species to the species select screen
        return sg
    entry = SpeciesGroupEntry(s['difficulty_priority'], s['enum'])
    sg[difficulty].species.append(entry)
    return sg


def generate_species_groups(sg):
    out = ''
    for name, group in sg.items():
        out += SPECIES_GROUP_TEMPLATE.format(
            name = name,
            position = group.position,
            width = group.width,
            species = ', '.join(
                e.enum for e in reversed(sorted(group.species))),
        )
    return out


def generate_species_type_data(s):
    if s['create_enum'] == False:
        return ''
    else:
        return '    %s,\n' % s['enum']


def maybe_write(filename, text):
    """Write `text` to `filename`, but only if the file would be created or changed"""
    if os.path.exists(filename):
        with open(filename, 'r') as f:
            cur = f.read()
        if cur == text:
            return

    with open(filename, 'w') as f:
        f.write(text)


def main():
    parser = argparse.ArgumentParser(description='Generate species-data.h')
    parser.add_argument('datadir', help='dat/species source dir')
    parser.add_argument('templatedir',
                    help='util/species-gen template source dir')
    parser.add_argument('species_data', help='species-data.h output file path')
    parser.add_argument('aptitudes', help='aptitudes.h output file path')
    parser.add_argument('species_groups',
                    help='species-groups.h output file path')
    parser.add_argument('species_type', help='species-type.h output file path')
    args = parser.parse_args()

    # Validate args
    if not os.path.isdir(args.datadir):
        print('datadir isn\'t a directory')
        sys.exit(1)
    if not os.path.isdir(args.templatedir):
        print('templatedir isn\'t a directory')
        sys.exit(1)

    # Load all species
    all_species = []
    for f_name in sorted(os.listdir(args.datadir)):
        if not f_name.endswith('.yaml'):
            continue
        f_path = os.path.join(args.datadir, f_name)
        try:
            species_spec = yaml.safe_load(open(f_path))
        except yaml.YAMLError as e:
            print("Failed to load %s: %s" % (f_name, e))
            sys.exit(1)

        try:
            species = Species(species_spec)
        except (ValueError, KeyError) as e:
            print("Failed to load %s" % f_name)
            traceback.print_exc()
            sys.exit(1)
        all_species.append(species)

    # Generate code
    species_data_out_text = load_template(args.templatedir,
                                                'species-data-header.txt')
    aptitudes_out_text = load_template(args.templatedir, 'aptitudes-header.txt')
    species_type_out_text = load_template(args.templatedir,
                                                'species-type-header.txt')

    species_data_template = load_template(args.templatedir,
                                                'species-data-species.txt')
    aptitude_template = load_template(args.templatedir, 'aptitude-species.txt')
    species_groups = SPECIES_GROUPS_TEMPLATE

    # Determine hard-coded enum order from species-type-header.txt
    enum_order = []
    header_lines = species_type_out_text.splitlines()
    for ln in header_lines:
        #print(ln)
        trimmed = ln.lstrip()
        if trimmed.startswith("SP_"):
            enum_name = re.split(',|/s+', trimmed)[0]
            if enum_name == "SP_FIRST_NONBASE_DRACONIAN" or enum_name == "SP_LAST_NONBASE_DRACONIAN":
                continue
            enum_order.append(enum_name)
            #print(enum_name)

    sorted_species = []
    for enum_name in enum_order:
        for species in all_species:
            if species['enum'] == enum_name:
                sorted_species.append(species)
                #print("Appended " + enum_name)
                break

    for species in sorted_species:
        # species-data.h
        species_data_out_text += species_data_template.format(**species)
        # aptitudes.h
        aptitudes_out_text += generate_aptitudes_data(species,
                                                            aptitude_template)
        # species-type.h
        species_type_out_text += generate_species_type_data(species)
        # species-groups.h
        species_groups = update_species_group(species_groups, species)

    species_data_out_text += load_template(args.templatedir,
                                        'species-data-footer.txt')
    maybe_write(args.species_data, species_data_out_text)

    aptitudes_out_text += load_template(args.templatedir,
                                        'aptitudes-footer.txt')
    maybe_write(args.aptitudes, aptitudes_out_text)

    species_type_out_text += load_template(args.templatedir,
                                        'species-type-footer.txt')
    maybe_write(args.species_type, species_type_out_text)

    species_groups_out_text = ''
    species_groups_out_text += load_template(args.templatedir,
                                        'species-groups-header.txt')
    species_groups_out_text += generate_species_groups(species_groups)
    species_groups_out_text += load_template(args.templatedir,
                                        'species-groups-footer.txt')
    maybe_write(args.species_groups, species_groups_out_text)


if __name__ == '__main__':
    main()



import errno
import logging
import os.path
import re
import smtplib
import time
import itertools
import functools
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import tornado.ioloop
import tornado.template
import tornado.gen

from webtiles import config

try:
    from typing import Any, Callable, Dict, Optional
    from typing.IO import TextIO
except ImportError:
    pass


# utility code for explicit debugging of potentially blocking I/O. We use this
# because while asyncio debugging is good at detecting when a blocking call
# happens, it is terrible at identifying what tornado 6+ is actually doing at
# the time.


class SlowWarning(object):
    def __init__(self, desc, time=None):
        if time is None:
            time = config.get('slow_io_alert')
        self.time = time
        self.desc = desc

    def __enter__(self):
        if self.time:
            self.start = time.monotonic()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.time:
            duration = time.monotonic() - self.start
            if duration >= self.time:
                logger = logging.getLogger("server.py")
                logger.warning('%s: %.3fs', self.desc, duration)
        return False


# backwards compatibility for older python, older Tornado. Also, future-proofing
# for internal API calls that could change. This check disables the slow
# callback diagnostic code if the tries fail.
_asyncio_available = True
try:
    def f():
        pass
    f.__qualname__ # for good measure..
    from asyncio.base_events import _format_handle
    import asyncio.events
    import reprlib
except:
    _asyncio_available = False

_aiofiles_available = False
try:
    import aiofiles
    _aiofiles_available = True
except:
    pass


@tornado.gen.coroutine
def open_and_write(filename, s):
    if _aiofiles_available:
        try:
            # jump through some hoops to not use async syntax, hopefully this
            # will not really be necessary for much longer...
            f = yield aiofiles.open(filename, mode='w').__aenter__()
            # is there a way of logging if the write thread gets bogged down?
            yield f.write(s)
        finally:
            yield f.close()
    else:
        with SlowWarning("Slow IO: write '%s'" % filename):
            with open(filename, "w") as f:
                f.write(s)


def func_repr(func):
    # from asyncio.format_helps._format_callback, not used for rendering
    if hasattr(func, '__qualname__') and func.__qualname__:
        return func.__qualname__
    elif hasattr(func, '__name__') and func.__name__:
        return func.__name__
    else:
        return repr(func)

def callback_arg_repr(f):
    # strip any instances of vacuous functools.partial, for readability (there
    # are a lot of these in practice)
    if isinstance(f, functools.partial) and not f.args:
        f = f.func
    return reprlib.repr(f)

_original_asyncio_run = None

# this function originally based on code from aiodebug:
#   https://gitlab.com/quantlane/libs/aiodebug
# License: Apache 2.0
def set_slow_callback_logging(slow_duration):
    global _asyncio_available
    if not _asyncio_available:
        return

    from webtiles import ws_handler

    global _original_asyncio_run
    if not slow_duration:
        if _original_asyncio_run:
            # de-monkeypatch
            asyncio.events.Handle._run = _original_asyncio_run
            _original_asyncio_run = None
        return

    # regular asyncio.Handle objects unfortunately delete most of their useful
    # info when they are canceled. When debugging a blocking canceled Handle
    # (which can definitely happen!) the best workaround I've found is to
    # use the following fairly insane monkeypatch, adding any extra info
    # to what is cached as needed, and printing it below. This is a huge
    # memory leak if used for anything other than debugging. (Note that
    # handle instances are largely read-only, hence the cache here.)

    # old_cancel = asyncio.Handle.cancel
    # canceled_handles = {}
    # def mpcancel(self):
    #     from tornado.platform.asyncio import BaseAsyncIOLoop
    #     old_repr_info = ["Cancelled"] + self._repr_info()
    #     old_cancel(self)
    #     canceled_handles[self] = old_repr_info

    # asyncio.Handle.cancel = mpcancel

    # the default repr call for asyncio Handles will limit argument strings
    # substantially (default, 20 chars) via reprlib, making it impossible to
    # get any useful info.
    reprlib.aRepr.maxother = 1000

    def on_slow_callback(name, duration):
        if name is None:
            return
        logger = logging.getLogger("server.py")
        logger.warning('Slow callback (%.3fs): %s', duration, name)
        global last_blocking_description
        last_blocking_description = "None"

    def format_tornado_handle(h):
        from tornado.platform.asyncio import BaseAsyncIOLoop

        # to see only certain callbacks when debugging, you can add a filter here.

        # the code here uses a certain level of undocumented / private APIs,
        # and ad-hoc introspection, so do try/except wrapping for future
        # proofing
        try:
            if (h._callback is not None
                            and func_repr(h._callback).startswith("IOLoop") # can this be done with introspection?
                            and h._args):
                # simplify down IOLoop callbacks to just the part that matters.
                # this will typically strip off something like:
                # `<TimerHandle when=X IOLoop._run_callback(functools.partial(...))>`
                # based on asyncio.format_helpers._format_args_and_kwargs
                # possibly items is always size 1 for an IOLoop callback?
                items = [callback_arg_repr(arg) for arg in h._args]
                s = 'IOLoop callback: {}'.format(', '.join(items))
                if h._cancelled:
                    s += " cancelled!"
                return s
            elif (h._callback is not None
                        # check for bound instance method: I cannot for the life of
                        # me figure out how to directly get the `method` type
                        and getattr(h._callback, '__self__', None)
                        and isinstance(h._callback.__self__, BaseAsyncIOLoop)
                        and getattr(h._callback, '__func__', None)
                        and h._callback.__func__ == BaseAsyncIOLoop._handle_events):
                # for event handlers, h._callback is a bound method instance on
                # BaseAsyncIOLoop, which is not very informative. Instead, find
                # the actual callback that Tornado used by introspecting on the
                # method object.
                s = "BaseAsyncIOLoop handler: {}".format(
                    repr(h._callback.__self__.handlers[h._args[0]]))
                if h._cancelled:
                    s += " cancelled!"
                return s
        except:
            pass

        # otherwise, fallback on the default (with the reprlib tweak). The
        # existence of this asyncio.base_events internal function is already
        # checked by a global try..except block.
        return _format_handle(h)

    if not _original_asyncio_run:
        _original_asyncio_run = asyncio.events.Handle._run

    def instrumented(self):
        global _original_asyncio_run
        t0 = time.monotonic()
        return_value = _original_asyncio_run(self)
        dt = time.monotonic() - t0
        if dt >= slow_duration:
            on_slow_callback(format_tornado_handle(self), dt)
        return return_value

    asyncio.events.Handle._run = instrumented


class TornadoFilter(logging.Filter):
    def filter(self, record):  # noqa A003: ignore shadowing builtin
        if record.module == "web" and record.levelno <= logging.INFO:
            return False
        return True


class DynamicTemplateLoader(tornado.template.Loader):  # type: ignore
    def __init__(self, root_dir):  # type: (str) -> None
        tornado.template.Loader.__init__(self, root_dir)

    def load(self, name, parent_path=None):
        name = self.resolve_path(name, parent_path=parent_path)
        # Fall back to template.html.default if template.html does not exist
        if not os.path.isfile(os.path.join(self.root, name)):
            name = name + '.default'

        if name in self.templates:
            template = self.templates[name]
            path = os.path.join(self.root, name)
            if os.path.getmtime(path) > template.load_time:  # type: ignore
                del self.templates[name]
            else:
                return template

        parent_cls = super(DynamicTemplateLoader, self)
        template = parent_cls.load(name, parent_path)  # type: ignore
        template.load_time = time.time()  # type: ignore
        return template

    _instances = {}  # type: Dict[str, DynamicTemplateLoader]

    @classmethod
    def get(cls, path):  # type: (str) -> DynamicTemplateLoader
        if path in cls._instances:
            return cls._instances[path]
        else:
            loader = DynamicTemplateLoader(path)
            cls._instances[path] = loader
            return loader


# light subclass to improve logging for the FileTailer scheduler
class PeriodicCallback(tornado.ioloop.PeriodicCallback):
    def __init__(self, check, interval_ms, source_desc=None):
        super(PeriodicCallback, self).__init__(check, interval_ms)
        self._source_desc = source_desc if source_desc else repr(check)

    def __repr__(self):
        return "PeriodicCallback(%s)" % self._source_desc


class FileTailer(object):
    def __init__(self, filename, callback):
        # type: (str, Callable[[str], Any], int) -> None
        self.file = None  # type: Optional[TextIO]
        self.filename = filename
        self.callback = callback
        self.scheduler = PeriodicCallback(self.check,
                config.get('milestone_interval'),
                "FileTailer('%s').check" % filename)
        self.scheduler.start()

    def check(self):  # type: () -> None
        if self.file is None:
            try:
                with SlowWarning("Slow IO: open '%s'" % self.filename):
                    self.file = open(self.filename, "r")
            except (IOError, OSError) as e:  # noqa
                if e.errno == errno.ENOENT:
                    return
                else:
                    raise

            with SlowWarning("Slow IO: seek '%s'" % self.filename):
                self.file.seek(os.path.getsize(self.filename))

        while True:
            pos = self.file.tell()
            with SlowWarning("Slow IO: readline '%s'" % self.filename):
                line = self.file.readline()
            if line.endswith("\n"):
                self.callback(line)
            else:
                self.file.seek(pos)
                return

    def stop(self):  # type: () -> None
        self.scheduler.stop()


_WHERE_ENTRY_REGEX = re.compile("(?<=[^:]):(?=[^:])")


def parse_where_data(data):  # type: (str) -> Dict[str, str]
    where = {}

    for entry in _WHERE_ENTRY_REGEX.split(data):
        if not entry.strip():
            continue
        field, _, value = entry.partition("=")
        where[field.strip()] = value.strip().replace("::", ":")
    return where


def send_email(to_address, subject, body_plaintext, body_html):
    # type: (str, str, str, str) -> None
    if not to_address:
        return

    logging.info("Sending email to '%s' with subject '%s'", to_address, subject)
    connected = False
    try:
        # establish connection
        # TODO: this should not be a blocking call at all...
        if config.get('smtp_use_ssl'):
            email_server = smtplib.SMTP_SSL(
                config.get('smtp_host'), config.get('smtp_port'))  # type: smtplib.SMTP
        else:
            email_server = smtplib.SMTP(config.get('smtp_host'), config.get('smtp_port'))
        connected = True

        # authenticate
        if config.get('smtp_user'):
            email_server.login(config.get('smtp_user'), config.get('smtp_password'))

        # build multipart message
        msg = MIMEMultipart('alternative')
        msg['Subject'] = subject
        msg['From'] = config.get('smtp_from_addr')
        msg['To'] = to_address

        part1 = MIMEText(body_plaintext, 'plain')
        part2 = MIMEText(body_html, 'html')

        msg.attach(part1)
        msg.attach(part2)

        # send
        email_server.sendmail(config.get('smtp_from_addr'), to_address, msg.as_string())
    finally:
        # end connection
        if connected:
            email_server.quit()


def validate_email_address(address):  # type: (str) -> Optional[str]
    # Returns an error string describing the problem, or None.
    # NOTE: a blank email address validates successfully.
    if not address:
        return None
    if " " in address:
        return "Email address can't contain a space"
    if "@" not in address:
        return "Expected email address to contain the @ symbol"
    if not re.match(r"[^@]+@[^@]+\.[^@]+", address):
        return "Invalid email address"
    if len(address) >= 80:
        return "Email address can't be more than 80 characters"
    return None


def humanise_bytes(num):  # type: (int) -> str
    """Convert raw bytes into human-friendly measure."""
    # TODO: Replace with pypi/humanize when we remove python 2 support.
    units = ["kilobytes", "megabytes", "gigabytes"]
    for index, unit in reversed(list(enumerate(units))):
        index += 1
        n = float(num) / 1000**index
        if n > 1:
            return "{} {}".format(round(n, 1), unit)
    return "{} bytes".format(num)



import os.path

import tornado
import tornado.web

from webtiles import config

try:
    from typing import Dict
except:
    pass

class GameDataHandler(tornado.web.StaticFileHandler):
    # async def _execute(self, transforms, *args, **kwargs):
    #     await tornado.web.StaticFileHandler._execute(self, transforms, *args, **kwargs)

    def initialize(self):
        if tornado.version_info[0] < 3:
            # ugly extreme backward compatibility hack; can hopefully be removed
            # once tornado 2.4 is out of the picture.
            super(GameDataHandler, self).initialize(".")
            self.root = "/" # just override the superclass for these old versions...
        else:
            super(GameDataHandler, self).initialize("/")

    def parse_url_path(self, url_path):
        # the path should already match "([0-9a-f]*\/.*)", from server.py
        import sys
        version, url_path = url_path.split("/", 1)
        if version not in GameDataHandler._client_paths:
            raise tornado.web.HTTPError(404)
        return super(GameDataHandler, self).parse_url_path(
                        GameDataHandler._client_paths[version] + "/" + url_path)

    def set_extra_headers(self, path):
        if config.get('game_data_no_cache'):
            self.set_header("Cache-Control",
                            "no-cache, no-store, must-revalidate")
            self.set_header("Pragma", "no-cache")
            self.set_header("Expires", "0")

    _client_paths = {} # type: Dict[str, str]

    @classmethod
    def add_version(cls, version, path):
        cls._client_paths[version] = os.path.abspath(path)



#!/usr/bin/env python3
"""
Generate the files job-data.h, job-type.h and job-groups.h from the YAML files
in ../dat/jobs/.
"""

from __future__ import print_function

import argparse
import os
import sys
import traceback
import collections
import re

if sys.version_info.major == 2:
    from collections import MutableMapping
else:
    from collections.abc import MutableMapping

import yaml  # pip install pyyaml

from genutil import *

class Job(MutableMapping):
    """Parser for YAML definition files.

    If any YAML content is invalid, the relevant parser function below should
    raise ValueError.
    """

    # TODO: unify with processing in from_yaml
    YAML_MAIN_FIELDS = {'TAG_MAJOR_VERSION', 'category', 'category_priority',
            'recommended_species', 'enum', 'name', 'short_name', 'str', 'int',
            'dex', 'weapon_choice', 'create_enum', 'skills', 'equipment',
            'spells'}

    def __init__(self, yaml_dict):
        self.backing_dict = dict()
        self.from_yaml(yaml_dict)

    def __getitem__(self, key):
        return self.backing_dict[key]

    def __setitem__(self, key, val):
        self.backing_dict[key] = val

    def __delitem__(self, key):
        del self.backing_dict[key]

    def __iter__(self):
        return iter(self.backing_dict)

    def __len__(self):
        return len(self.store)

    def print_unknown_warnings(self, s):
        for key in s:
            if key not in self.YAML_MAIN_FIELDS:
                print("species_gen.py warning: Unknown field '%s' in species %s"
                                    % (key, self['enum']), file=sys.stderr)

    def from_yaml(self, s):
        # Pre-validation
        if s.get('TAG_MAJOR_VERSION', None) is not None:
            if not isinstance(s['TAG_MAJOR_VERSION'], int):
                raise ValueError('TAG_MAJOR_VERSION must be an integer')
        self.starting_job = s.get('difficulty') != False
        has_recommended_species = bool(s.get('recommended_species'))
        if self.starting_job != has_recommended_species:
            raise ValueError('recommended_jobs must not be empty (or'
                                                ' category must be False)')

        # Set attributes
        self['enum'] = validate_string(s['enum'], 'enum', 'JOB_[A-Z_]+$')
        self['name'] = validate_string(s['name'], 'name', '..+')
        self['short_name'] = s.get('short_name', s['name'][:2])
        # TODO: currently done by skill enum values; change to be strings?
        self['skills'] = skills(s.get('skills', {}))
        self['str'] = validate_int_range(s['str'], 'str', -15, 15)
        self['int'] = validate_int_range(s['int'], 'int', -15, 15)
        self['dex'] = validate_int_range(s['dex'], 'dex', -15, 15)
        self['recommended_species'] = recommended_species(
                                            s.get('recommended_species', []))
        self['spells'] = spells(s.get('spells', []))
        self['equipment'] = equipment(s.get('equipment', []))
        self['weapon_choice'] = weapon_choice(
                                        s.get('weapon_choice', 'none'))
        self['category'] = category(s.get('category'))
        self['category_priority'] = validate_int_range(category_priority(
            s.get('category_priority', 0)), 'category_priority', 0, 1000)
        self['create_enum'] = validate_bool(
                                    s.get('create_enum', True), 'create_enum')

        if 'TAG_MAJOR_VERSION' in s:
            self['tag_major_version_opener'] = (
                        "#if TAG_MAJOR_VERSION == %s" % s['TAG_MAJOR_VERSION'])
            self['tag_major_version_closer'] = "#endif"
        else:
            self['tag_major_version_opener'] = ''
            self['tag_major_version_closer'] = ''
        self.print_unknown_warnings(s)

JobGroup = collections.namedtuple('JobGroup', ['position', 'width', 'jobs'])
JobGroupEntry = collections.namedtuple('JobGroupEntry', ['priority', 'enum'])
JOB_GROUPS_TEMPLATE = {
    'Warrior': JobGroup('coord_def(0, 0)', '15', []),
    'Adventurer': JobGroup('coord_def(0, 7)', '15', []),
    'Zealot': JobGroup('coord_def(15, 0)', '20', []),
    'Warrior-mage': JobGroup('coord_def(35, 0)', '21', []),
    'Mage': JobGroup('coord_def(56, 0)', '22', []),
}
JOB_GROUP_TEMPLATE = """
    {{
        "{name}",
        {position},
        {width},
        {{ {jobs} }}
    }},
"""

# SK_WEAPON is used when weapon is chosen by player
ALL_SKILLS = ('fighting', 'short blades', 'long blades', 'axes',
    'maces and flails', 'polearms', 'staves', 'ranged weapons',
    'throwing', 'armour', 'dodging', 'stealth',
    'shields', 'unarmed combat', 'spellcasting', 'conjurations',
    'hexes', 'summonings', 'necromancy', 'translocations',
    'fire magic', 'ice magic', 'air magic', 'earth magic',
    'alchemy', 'invocations', 'evocations', 'shapeshifting', 'forgecraft',
    'weapon')
ALL_WCHOICES = ('none', 'plain', 'good')

#SPELL_ENUM = 'SPELL'
SKILL_ENUM = 'SK'
SPECIES_ENUM = 'SP'

def recommended_species(species):
     # Allow writing 'draconian' for SP_BASE_DRACONIAN
    out = []
    for race in species:
        if race.lower() == "draconian":
            race = "base draconian"
        out.append(enumify(validate_string(race, 'Species',
                                           '[A-Za-z_ ]+'), SPECIES_ENUM))
    return make_list(', '.join(out))


def skills(sk):
    out = []
    for skill, val in sorted(sk.items()):
        if skill not in ALL_SKILLS:
            raise ValueError("Unknown skill (typo?): %s" % skill)
        # Hack: allow "maces and flails" for SK_MACES_FLAILS
        if skill.lower() == "maces and flails":
            skill = "maces flails"
        validate_int_range(val, skill, 0, 10)
        out.append("{{ {skill}, {val} }}".format(
                            skill=enumify(skill, SKILL_ENUM), val=val))
    return make_list(', '.join(out))


def equipment(items):
    return make_list(', '.join('"{item}"'.format(item=item) for item in items))


def weapon_choice(wchoice):
    if wchoice not in ALL_WCHOICES:
        raise ValueError("Unknown weapon choice: %s" % wchoice)
    return wchoice

def spells(spell_list):
    return make_list(',\n'.join('        {spell}'.format(spell=spell)
                                for spell in spell_list), True)


def category(c):
    if c not in JOB_GROUPS_TEMPLATE.keys() and c is not False:
        raise ValueError("Unknown category: %s" % c)
    return c


def category_priority(prio):
    try:
        return int(prio)
    except ValueError:
        raise ValueError('category_priority value "%s" is not an integer' %
                                prio)


def update_job_group(sg, s):
    category = s['category']
    if category is False:
        # Don't add this species to the species select screen
        return sg
    entry = JobGroupEntry(s['category_priority'], s['enum'])
    sg[category].jobs.append(entry)
    return sg


def generate_job_groups(sg):
    out = ''
    for name, group in sg.items():
        out += JOB_GROUP_TEMPLATE.format(
            name = name,
            position = group.position,
            width = group.width,
            jobs = ', '.join(
                e.enum for e in reversed(sorted(group.jobs))),
        )
    return out


def generate_job_type_data(s):
    if s['create_enum'] == False:
        return ''
    else:
        return '    %s,\n' % s['enum']


def maybe_write(filename, text):
    """Write `text` to `filename`, but only if the file would be created or changed"""
    if os.path.exists(filename):
        with open(filename, 'r') as f:
            cur = f.read()
        if cur == text:
            return

    with open(filename, 'w') as f:
        f.write(text)


def main():
    parser = argparse.ArgumentParser(description='Generate job-data.h')
    parser.add_argument('datadir', help='dat/jobs source dir')
    parser.add_argument('templatedir',
                    help='util/job-gen template source dir')
    parser.add_argument('job_data', help='job-data.h output file path')
    parser.add_argument('job_groups', help='job-groups.h output file path')
    parser.add_argument('job_type', help='job-type.h output file path')
    args = parser.parse_args()

    # Validate args
    if not os.path.isdir(args.datadir):
        print('datadir isn\'t a directory')
        sys.exit(1)
    if not os.path.isdir(args.templatedir):
        print('templatedir isn\'t a directory')
        sys.exit(1)

    # Load all species
    all_jobs = []
    for f_name in sorted(os.listdir(args.datadir)):
        if not f_name.endswith('.yaml'):
            continue
        f_path = os.path.join(args.datadir, f_name)
        try:
            job_spec = yaml.safe_load(open(f_path))
        except yaml.YAMLError as e:
            print("Failed to load %s: %s" % (f_name, e))
            sys.exit(1)

        try:
            job = Job(job_spec)
        except (ValueError, KeyError) as e:
            print("Failed to load %s" % f_name)
            traceback.print_exc()
            sys.exit(1)
        all_jobs.append(job)

    # Generate code
    job_data_out_text = load_template(args.templatedir,
                                                'job-data-header.txt')
    job_type_out_text = load_template(args.templatedir,
                                                'job-type-header.txt')

    job_data_template = load_template(args.templatedir,
                                                'job-data-jobs.txt')

    job_groups = JOB_GROUPS_TEMPLATE
    for job in all_jobs:
        # job-data.h
        job_data_out_text += job_data_template.format(**job)
        # job-type.h
        job_type_out_text += generate_job_type_data(job)
        # job-groups.h
        job_groups = update_job_group(job_groups, job)

    job_data_out_text += load_template(args.templatedir,
                                        'job-data-deprecated-jobs.txt')
    job_data_out_text += load_template(args.templatedir,
                                        'job-data-footer.txt')
    maybe_write(args.job_data, job_data_out_text)

    job_type_out_text += load_template(args.templatedir,
                                        'job-type-footer.txt')
    maybe_write(args.job_type, job_type_out_text)

    job_groups_out_text = ''
    job_groups_out_text += load_template(args.templatedir,
                                        'job-groups-header.txt')
    job_groups_out_text += generate_job_groups(job_groups)
    job_groups_out_text += load_template(args.templatedir,
                                        'job-groups-footer.txt')
    maybe_write(args.job_groups, job_groups_out_text)


if __name__ == '__main__':
    main()



import pytest

from webtiles import util


class Test_validate_email_address:
    @pytest.mark.parametrize("email, valid", [
        ("user@gmail.com", True),
        ("user.name@gmail.domain.com", True),
        ("", True),  # !!!
        ("user name@gmail.com", False),
        ("user", False),
        ("user@domain", False),
        ("u" * 80 + "@gmail.com", False),
    ])
    def test_validate_email_address(self, email, valid):
        result = util.validate_email_address(email)
        if valid:
            assert result is None
        else:
            assert result is not None


class Test_humanise_bytes:
    @pytest.mark.parametrize("num, expected", [
        (1, "1 bytes"),
        (1000, "1000 bytes"),
        (1000000, "1000.0 kilobytes"),
        (1000000000, "1000.0 megabytes"),
        (1000000000000, "1000.0 gigabytes"),
    ])
    def test_humanise_bytes(self, num, expected):
        assert util.humanise_bytes(num) == expected



#!/usr/bin/env python
from __future__ import absolute_import

import os

import webtiles, webtiles.server, webtiles.config

def main():
    import config as server_config
    webtiles.config.init_config_from_module(server_config)
    webtiles.config.server_path = os.path.dirname(os.path.abspath(__file__))
    webtiles.server.run_util()

if __name__ == "__main__":
    main()



import asyncio
import fcntl
import os
import pty
import resource
import signal
import struct
import sys
import termios
import time

import tornado.ioloop
from tornado.escape import to_unicode
from tornado.ioloop import IOLoop

from webtiles import util, config

BUFSIZ = 2048

class TerminalRecorder(object):
    def __init__(self,
                 command, # type: List[str]
                 logger,
                 termsize,
                 env_vars, # type: Dict[str, str]
                 game_cwd, # type: Optional[str]
                 ):
        """
        Args:
            command: argv of command to run, eg [cmd, args, ...]
            env_vars: dictionary of environment variables to set. The variables
                COLUMNS, LINES, and TERM cannot be overridden.
        """
        self.command = command
        self.ttyrec = None
        self.desc = "TerminalRecorder"
        self.returncode = None
        self.output_buffer = b""
        self.termsize = termsize
        self.env_vars = env_vars
        self.game_cwd = game_cwd

        self.pid = None
        self.child_fd = None

        self.end_callback = None
        self.output_callback = None
        self.activity_callback = None
        self.error_callback = None

        self.errpipe_read = None
        self.error_buffer = b""

        self.logger = logger

    def start(self, ttyrec_filename, id_header):
        if ttyrec_filename:
            self.desc = ttyrec_filename
            with util.SlowWarning("Slow IO: open '%s'" % self.desc):
                self.ttyrec = open(ttyrec_filename, "wb") # type: Optional[BinaryIO]
        if id_header:
            self.write_ttyrec_chunk(id_header, flush=True)

        self._spawn()

    def is_started(self):
        return self.pid is not None and self.pid != 0

    def _spawn(self):
        self.errpipe_read, errpipe_write = os.pipe()

        self.pid, self.child_fd = pty.fork()

        if self.pid == 0:
            # We're the child

            # attempt to avoid some race conditions in global effects that can
            # be triggered by early signals (before execvpe replaces process
            # state)
            config.set("pidfile", None)

            # remove server's signal handling for the interim.
            # Note: I haven't found a reliable way to do this, because shared
            # resources are copied on fork. See:
            #   https://github.com/python/cpython/issues/66197
            # so, sometimes, a race condition here leads to the main process
            # reloading its config...currently some key cases are handled
            # elsewhere by a brute force delay.
            asyncio.get_event_loop().remove_signal_handler(signal.SIGHUP)

            # Set window size
            cols, lines = self.get_terminal_size()
            s = struct.pack("HHHH", lines, cols, 0, 0)
            fcntl.ioctl(sys.stdout.fileno(), termios.TIOCSWINSZ, s)

            os.close(self.errpipe_read)
            os.dup2(errpipe_write, 2)

            # Make sure not to retain any files from the parent
            max_fd = resource.getrlimit(resource.RLIMIT_NOFILE)[0]
            for i in range(3, max_fd):
                try:
                    os.close(i)
                except OSError:
                    pass

            # And exec
            env            = dict(os.environ)
            env.update(self.env_vars)
            env["COLUMNS"] = str(cols)
            env["LINES"]   = str(lines)
            env["TERM"]    = "linux"
            if self.game_cwd:
                os.chdir(self.game_cwd)
            try:
                os.execvpe(self.command[0], self.command, env)
            except OSError:
                sys.exit(1)

        # We're the parent
        os.close(errpipe_write)
        if not self.ttyrec:
            self.desc = "TerminalRecorder (fd %d)" % self.child_fd

        IOLoop.current().add_handler(self.child_fd,
                                     self._handle_read,
                                     IOLoop.ERROR | IOLoop.READ)

        IOLoop.current().add_handler(self.errpipe_read,
                                     self._handle_err_read,
                                     IOLoop.READ)

    def _handle_read(self, fd, events):
        if events & IOLoop.READ:
            try:
                with util.SlowWarning("Slow IO: os.read (session '%s')" % self.desc):
                    buf = os.read(fd, BUFSIZ)
            except (OSError, IOError):
                self.poll() # fd probably closed?
                return

            if len(buf) > 0:
                try:
                    self.write_ttyrec_chunk(buf)
                except OSError as e:
                    # should something more happen?
                    self.logger.error("Failed to write ttyrec chunk! (%s)" % e)

                if self.activity_callback:
                    self.activity_callback()

                self.output_buffer += buf
                self._do_output_callback()

            self.poll()

        if events & IOLoop.ERROR:
            self.poll()

    def _handle_err_read(self, fd, events):
        if events & IOLoop.READ:
            with util.SlowWarning("Slow IO: os.read stderr (session '%s')" % self.desc):
                buf = os.read(fd, BUFSIZ)

            if len(buf) > 0:
                self.error_buffer += buf
                self._log_error_output()

            self.poll()

    def write_ttyrec_header(self, sec, usec, l):
        if self.ttyrec is None:
            return
        s = struct.pack("<iii", sec, usec, l)
        self.ttyrec.write(s)

    def write_ttyrec_chunk(self, data, flush=False):
        if self.ttyrec is None:
            return
        with util.SlowWarning("Slow IO: write_ttyrec_chunk '%s'" % self.desc):
            t = time.time()
            self.write_ttyrec_header(int(t), int((t % 1) * 1000000), len(data))
            self.ttyrec.write(data)
        if flush:
            self.flush_ttyrec()

    def flush_ttyrec(self):
        if self.ttyrec is None:
            return
        with util.SlowWarning("Slow IO: flush '%s'" % self.desc):
            self.ttyrec.flush()

    def _do_output_callback(self):
        pos = self.output_buffer.find(b"\n")
        while pos >= 0:
            line = self.output_buffer[:pos]
            self.output_buffer = self.output_buffer[pos + 1:]

            if len(line) > 0:
                if line[-1] == b"\r": line = line[:-1]

                if self.output_callback:
                    self.output_callback(to_unicode(line))

            pos = self.output_buffer.find(b"\n")

    def _log_error_output(self):
        pos = self.error_buffer.find(b"\n")
        while pos >= 0:
            line = self.error_buffer[:pos]
            self.error_buffer = self.error_buffer[pos + 1:]

            if len(line) > 0:
                if line[-1] == b"\r": line = line[:-1]

                self.logger.info("ERR: %s", to_unicode(line))
                if self.error_callback:
                    self.error_callback(to_unicode(line))

            pos = self.error_buffer.find(b"\n")


    def send_signal(self, signal):
        if not self.is_started():
            raise RuntimeError("Can't send a signal without a child process to send it to!")
        os.kill(self.pid, signal)

    def poll(self):
        if self.returncode is None:
            pid, status = os.waitpid(self.pid, os.WNOHANG)
            if pid == self.pid:
                if os.WIFSIGNALED(status):
                    self.returncode = -os.WTERMSIG(status)
                elif os.WIFEXITED(status):
                    self.returncode = os.WEXITSTATUS(status)
                else:
                    # Should never happen
                    raise RuntimeError("Unknown child exit status!")

            if self.returncode is not None:
                IOLoop.current().remove_handler(self.child_fd)
                IOLoop.current().remove_handler(self.errpipe_read)

                os.close(self.child_fd)
                os.close(self.errpipe_read)

                if self.ttyrec:
                    with util.SlowWarning("Slow IO: close '%s'" % self.desc):
                        self.ttyrec.close()

                if self.end_callback:
                    self.end_callback()

                # accessed in the default end callback for logging
                self.pid = None

        return self.returncode

    def get_terminal_size(self):
        return self.termsize

    def write_input(self, data):
        if self.poll() is not None: return

        while len(data) > 0:
            written = os.write(self.child_fd, data)
            data = data[written:]



#!/usr/bin/env python3

"""
Works with both Python 2 & 3. If that changes, update how the Makefile calls
this.
"""

from subprocess import Popen, PIPE

import os
import shlex
import sys

if len(sys.argv) < 2:
    sys.stderr.write('CXX not specified\n')
    sys.exit(1)

# Not just the name of the compiler, it can have arguments to
CXX = sys.argv[1]
command = shlex.split(CXX) + ['-c', 'conftest.cc']

out = open('conftest.cc', 'w')
out.write("""
#include <unistd.h>
int main()
{
    fdatasync(1);
    return 0;
}
""")
out.close()

process = Popen(command, stdout=PIPE, stderr=PIPE)
(output, err) = process.communicate()
exit_code = process.wait()
FDATASYNC = "#define CRAWL_HAVE_FDATASYNC" if exit_code == 0 else "#undef CRAWL_HAVE_FDATASYNC"

out = open('conftest.cc', 'w')
out.write("""
#include <cstring>
using namespace std;
int main()
{
    const char *src = "hello";
    char dst[10];
    strlcpy(dst, src, sizeof(dst));
    return 0;
}
""")
out.close()

process = Popen(command, stdout=PIPE, stderr=PIPE)
(output, err) = process.communicate()
exit_code = process.wait()
STRLCPY = "#define CRAWL_HAVE_STRLCPY" if exit_code == 0 else "#undef CRAWL_HAVE_STRLCPY"

out = open('conftest.cc', 'w')
out.write("""
#include <cstdlib>
using namespace std;
int main()
{
    char file[] = "XXXXXX";
    mkstemp(file);
    return 0;
}
""")
out.close()

process = Popen(command, stdout=PIPE, stderr=PIPE)
(output, err) = process.communicate()
exit_code = process.wait()
MKSTEMP = "#define CRAWL_HAVE_MKSTEMP" if exit_code == 0 else "#undef CRAWL_HAVE_MKSTEMP"

out = open('conftest.cc', 'w')
out.write("""
#include <unistd.h>
int main()
{
    usleep(0);
    return 0;
}
""")
out.close()

process = Popen(command, stdout=PIPE, stderr=PIPE)
(output, err) = process.communicate()
exit_code = process.wait()
USLEEP = "#define CRAWL_HAVE_USLEEP" if exit_code == 0 else "#undef CRAWL_HAVE_USLEEP"

try:
    os.remove('conftest.cc')
except OSError:
    pass

try:
    os.remove('conftest.o')
except OSError:
    pass

try:
    os.remove('conftest.obj')
except OSError:
    pass

output_contents = """#pragma once

// Generated by util/configure.py

"""
output_contents += FDATASYNC
output_contents += '\n'
output_contents += STRLCPY
output_contents += '\n'
output_contents += MKSTEMP
output_contents += '\n'
output_contents += USLEEP
output_contents += '\n'

contents_match = False
try:
    config_file = open('config.h', 'r')
    current_contents = config_file.read()
    config_file.close()
    contents_match = current_contents == output_contents
except OSError:
    pass

if not contents_match:
    out = open('config.h', 'w')
    out.write(output_contents)
    out.close()



# utilities for dealing with webtiles configuration. The actual configuration
# data does *not* go in here.

import builtins
import collections
import itertools
import os.path
import logging
import re
import sys
import yaml
try:
    from collections.abc import MutableMapping
except:
    from collections import MutableMapping

from webtiles import load_games, bans

server_config = {}
source_file = None
source_module = None
server_path = None

# "temporary" compatibility shim for config calls in templates
allow_password_reset = False
admin_password_reset = False

# rudimentary code for log msgs before server.py:init_logging is called
# probably better just to avoid using as much as possible rather than to make
# it any more sophisticated. (Or, refactor early server setup so it's not
# needed..)
_early_logging = []


def early_log(s):
    global _early_logging
    _early_logging.append(s)

def do_early_logging():
    global _early_logging
    for s in _early_logging:
        logging.info(s)
    _early_logging = []


# game definitions and configuration
# note: get('games') should not be used to get game data! That call would get
# only whatever games are defined in the config module.
games = collections.OrderedDict()
game_templates = collections.OrderedDict()
# see also `metatemplates`, defined below

# mapping of game keys to game mode names
game_modes = {}  # type: Dict[str, str]

# mapping of binary keys to game keys
binaries_to_games = {}  # type: Dict[str, str]

# default values for module-level get calls
# for values not in this dict, the default is None
defaults = {
    'dgl_mode': True,
    'logging_config': {
        "level": logging.INFO,
        "format": "%(asctime)s %(levelname)s: %(message)s"
    },
    'server_socket_path': None,
    'watch_socket_dirs': False,
    'milestone_file': [],
    'status_file_update_rate': 30,
    'lobby_update_rate': 2,
    'recording_term_size': (80, 24),
    'max_connections': 100,
    'connection_timeout': 10 * 60,
    'max_idle_time': 5 * 60 * 60,
    'max_lobby_idle_time': 3 * 60 * 60,
    'use_gzip': True,
    'kill_timeout': 10,
    'nick_regex': r"^[a-zA-Z0-9]{3,20}$",
    'max_passwd_length': 20,
    'allow_password_reset': False,
    'admin_password_reset': False,
    'crypt_algorithm': "broken", # should this be the default??
    'crypt_salt_length': 16,
    'login_token_lifetime': 7, # Days; set to <= 0 to disable
    'recovery_token_lifetime': 12, # hours
    'daemon': False,
    'development_mode': False,
    'no_cache': False,
    'live_debug': False,
    'lobby_update_rate': 2,
    'load_logging_rate': 0,
    'milestone_interval': 1000, # ms
    'slow_callback_alert': None,
    'slow_io_alert': 0.250,
    'games': collections.OrderedDict([]),
    'templates': collections.OrderedDict([]),
    'use_game_yaml': None, # default: load games.d if games is empty
    'banned': [],
    'bot_accounts': False,
    'wizard_accounts': False,
    'allow_anon_spectate': True,
    'enable_ttyrecs': True,
    'max_chat_length': 1000,
}


# light wrapper class that maps get/set/etc to getattr/setattr/etc
# doesn't bother to implement most of the dict interface...
class ConfigModuleWrapper(object):
    def __init__(self, module):
        self.module = module

        # default override location for config settings
        self._load_override_file(os.path.join(
                        self.get("server_path", ""), "config.yml"))

        # default override locations for ban lists. This first filename
        # is intentionally the same as the scoring db's ban list filename.
        load_banfile(os.path.join(
                        self.get("server_path", ""), "banned_players.yml"))
        load_banfile(os.path.join(
                        self.get("server_path", ""), "banned_players.txt"))

    def _load_override_file(self, path):
        if not os.path.isfile(path):
            return
        with open(path) as f:
            override_data = yaml.safe_load(f)
            if not isinstance(override_data, dict):
                raise ValueError("'%s' must be a map" % path)
            early_log("Loading config overrides from: '%s'" % path)
            for key, value in override_data.items():
                if key == 'games':
                    # right now this completely overwrites anything set in the
                    # module, so issue a warning
                    if self.get('games', {}):
                        early_log("Warning: overriding `games` with value from " + path)
                    # fallthrough
                elif key == 'banned':
                    # we probably don't want to overwrite any ban lists set
                    # in the config module
                    add_ban_list(value)
                    continue

                setattr(self.module, key, value)

    def get(self, key, default):
        return getattr(self.module, key, default)

    def __setitem__(self, key, val):
        setattr(self.module, key, val)

    def pop(self, key):
        r = getattr(self.module, key)
        delattr(self.module, key)
        return r

    def __contains__(self, key):
        return hasattr(self.module, key)


# classic config: everything is just done in a module
# TODO: allow this to be loaded directly from yml somehow? Right now, there
# always must be a config module, and then an override from yml is possible
def init_config_from_module(module):
    global server_config, source_file, source_module
    source_module = module
    server_config = ConfigModuleWrapper(module)
    source_file = os.path.abspath(module.__file__)
    global allow_password_reset, admin_password_reset
    allow_password_reset = get('allow_password_reset')
    admin_password_reset = get('admin_password_reset')


def init_config_timeouts():
    from webtiles import ws_handler, util
    # anything added here should be robust to reload()...
    util.set_slow_callback_logging(get('slow_callback_alert'))
    ws_handler.do_load_logging(True)


def reload_namespace_resets():
    # ensure that the ban list always gets rebuilt
    set('banned', [])


def reload():
    global source_file, source_module
    try:
        logging.warning("Reloading config from %s", source_file)
        from importlib import reload

        # major caveat: this will not reset the namespace before doing the
        # reload. So to return something to the default value requires an
        # explicit setting. XX is there anything better to do about this?
        reload_namespace_resets()
        reload(source_module)
        init_config_from_module(source_module)
        do_early_logging()
        init_config_timeouts()
        try:
            load_game_data(True)
        except ValueError:
            logging.error("Game data reload failed!", exc_info=True)
            # if you get to here, game data is probably messed up. But there's
            # nothing to be done, probably...
        # XX it might be good to revalidate here, but I'm not sure how to
        # recover from errors
    except:
        logging.error("Config reload failed!", exc_info=True)


class GameConfig(MutableMapping):
    def __init__(self, params, use_template=True, game_id=None):
        template_name = None
        if isinstance(params, GameConfig): # copy constructor, essentially
            # possibly should be a deep copy?
            self._store = params._store.copy()
            if not game_id:
                game_id = params.id
            template_name = params.template if use_template else None
            #self.template = params.template if use_template else None
        else:
            self._store = params
            # note: this prevents the `template` param from being modified via
            # the mapping interface
            # self.template = self._store.pop('template', 'default') if use_template else None

        if game_id:
            self.id = game_id
            self._store['id'] = game_id
        else:
            self.id = params['id'] # must be set
        if use_template and template_name is None and not is_metatemplate(self.id):
            template_name = self._store.pop('template', 'default')
        self.template = template_name

    def copy(self):
        return GameConfig(self)

    def __repr__(self):
        params = self._store.copy()
        extras = ""
        if self.template:
            params['template'] = self.template
        else:
            extras = ", use_template=False"
        return "GameConfig(%s%s)" % (repr(params), extras)

    def get_defaults(self):
        if self.template:
            return get_template(self.template)
        else:
            return {}

    def templated(self, key, username=None, default=""):
        val = self.get(key, default)
        if isinstance(val, str):
            return dgl_format_str(val, username, self)
        elif isinstance(val, (list, tuple)): # ugh, but we need this for pre_options
            return [dgl_format_str(e, username, self)
                    if isinstance(e, str) else e for e in val]
        else:
            return val

    def templated_dict(self, username):
        r = {}
        for k in self:
            r[k] = self.templated(k, username=username)
        return r

    def get_call_base(self):
        # these params can't be templated with a username...
        call = [self.templated("crawl_binary")]
        if "pre_options" in self:
            call += self.templated("pre_options")
        return call

    def get_binary_key(self):
        # return a value that is suitable for using as a hash key for a crawl
        # binary. This intentionally accommodates `pre_options`, because for
        # dgamelaunch-config setups, stable is called via a wrapper script
        # with the version set via `pre_options`, so this is necessary to
        # differentiate different stable versions.
        return " ".join(self.get_call_base())

    def validate(self):
        # check for loops in templating
        seen = builtins.set()
        t = self.get_defaults()
        while isinstance(t, GameConfig) and t.template:
            if t.id in seen:
                logging.error("Loop in template dependencies from game %s: %s", self.id, t.id)
                return False
            seen.add(t.id)
            t = t.get_defaults()

        # warn if this is a non-templatable GameConfig
        if 'template' in self._store:
            logging.warning("Ignoring template value '%s' in %s",
                                            self._store['template'], self.id)
        return True

    def validate_game(self):
        # check for templating errors in parameter values. We want to do this
        # only for full games, not templates. Note that %n isn't validated
        # here, but is checked against a hardcoded list of keys in the
        # validate_game_dict call.
        for k in self:
            try:
                self.templated(k, username="test")
            except:
                logging.error("Error when templating game %s, key %s: %s",
                    self.id, k, repr(self[k]), exc_info=True)
                return False

        if not self.validate():
            return False
        return load_games.validate_game_dict(self)

    def __getitem__(self, key):
        # the exception handling pattern here is so that error messages are
        # cleaner with multiple levels of templating. Otherwise, we get a bunch
        # of hard to parse `During handling ...` traces, and a messy final
        # trace. XX once we can 100% assume py3.3, this would be cleaner if
        # redone using `raise from None`.
        try:
            return self._store[key]
        except KeyError:
            pass
        try:
            return self.get_defaults()[key]
        except KeyError:
            pass
        raise KeyError(key)

    def __setitem__(self, key, val):
        self._store[key] = val

    def __delitem__(self, key):
        del self._store[key]

    def clear(self):
        # override the mixin behavior. This really does clear everything,
        # because it clears `template`.
        self._store.clear() # override

    def __iter__(self):
        return itertools.chain(
            iter(self._store),
            filter(lambda k: k not in self._store, iter(self.get_defaults())))

    def __len__(self):
        return len(self._store) + len([k for k in self.get_defaults() if not k in self._store])


# metatemplates that ground out recursion. `default` can be adjusted using
# the `define_default` function below, `base` should be left alone.
# outside this file, only modified in load_games.py:load_games
metatemplates = dict(
    base = GameConfig({}, use_template=False, game_id='base'),
    default = GameConfig({}, use_template=False, game_id='default')
    )


def is_metatemplate(k):
    global metatemplates
    return k in metatemplates


def define_default(params):
    global metatemplates
    new = GameConfig(params, use_template=False, game_id='default')
    if not new.validate():
        logging.warning("Ignoring invalid defaults")
        return False
    if len(metatemplates['default']) > 1 and new != metatemplates['default']:
        logging.warning("Replacing existing default game template")
    metatemplates['default'] = new
    return True


def get_template(k):
    global metatemplates, game_templates
    if k is None:
        k = 'default'
    if is_metatemplate(k):
        return metatemplates[k]
    else:
        return game_templates[k]


# templating for strings in game config
def dgl_format_str(s, username, game_params):  # type: (str, str, Any) -> str
    # XX this should probably implement %% as an escape

    def tsub(s, code, name, base, xform=lambda x: x):
        if s.find(code) >= 0:
            if base is None:
                raise ValueError("%s used in config templating but not set in config: %s" % (name, s))
            s = s.replace(code, xform(base))
        return s

    # sometimes this is called without a username available, so gracefully
    # ignore if it is not provided
    if username is not None:
        s = tsub(s, "%n", 'username', username)

    # provide the version info in various ways useful for constructing paths
    version = game_params.get('version')

    s = tsub(s, "%v", 'version', version)
    s = tsub(s, "%V", 'version', version, lambda x: x.capitalize())
    # %r is the version as a string that follows a "0." (if not found, the
    # entire version string is used). This is used for building directory
    # names on some dgl servers.
    s = tsub(s, "%r", 'version', version, lambda x: x.split("0.", 1)[-1]) # XX a bit brittle
    return s


# use this to get templated game config values
def game_param(game_key, param, username=None, default=""):
    global games
    game = games[game_key]
    return game.templated(param, username=username, default=default)


def get(key, default=None):
    global server_config, defaults
    return server_config.get(key, defaults.get(key, default))


# note: shadows type `set`, use `builtins.set` if you need the type name
def set(key, val):
    global server_config
    server_config[key] = val


def pop(key):
    global server_config
    return server_config.pop(key)


def has_key(key):
    global server_config
    return key in server_config


def check_keys_all(required, raise_on_missing=False):
    # accept either a single str, or an iterable for `required`
    if isinstance(required, str):
        required = [required]
    for k in required:
        if not has_key(k) or get(k) is None:
            if raise_on_missing:
                raise ValueError("Webtiles config: Missing configuration key: %s" % k)
            return False
    return True


def check_keys_any(required, raise_on_missing=False):
    # use `has_keys`: if any member of required is itself a list, require
    # all keys in the list
    if not any([check_keys_all(key) for key in required]):
        if raise_on_missing:
            raise ValueError("Webtiles config: Need at least one of %s!" %
                                        ", ".join([repr(r) for r in required]))
        return False
    return True


def check_game_config():
    success = True
    global games
    for game in games.values():
        if not os.path.exists(game["crawl_binary"]):
            logging.warning("Crawl executable for %s ('%s') doesn't exist!",
                            game.id, game["crawl_binary"])
            success = False

        if ("client_path" in game and
                        not os.path.exists(game["client_path"])):
            logging.warning("Client data path for %s ('%s') doesn't exist!",
                            game.id, game["client_path"])
            success = False

    return success


# aggregate check on a bunch of options values to determine whether we should
# load yaml files from games.d/
def using_games_dir():
    # backwards compatibility: deprecated games_config_dir was either None to
    # disable, or a str, but treat it now as just a bool-like. (Slightly
    # awkward as None has a different semantics for use_game_yaml.)
    if has_key('games_config_dir'):
        return bool(get('games_config_dir'))
    else:
        return get('use_game_yaml') # True, False, or None (to ignore)


def add_ban_list(ban_list):
    cur = get('banned')
    cur.extend(ban_list)
    # explicitly set, in case we started with the default
    set('banned', cur)


# check a username against various config options that may disallow it
def check_name(username):
    # note: it's possible that nick_regex should be handled differently, since
    # it is always set and used as a syntactic check; the others are used
    # check for bad names of some kind
    if not re.match(get('nick_regex'), username):
        return False
    if get('nick_check_fun') and not get('nick_check_fun')(username):
        return False
    return not bans.do_ban_check(username, get('banned'))


def load_banfile(path, require_exists=False):
    if require_exists and not os.path.isfile(path):
        raise ValueError("Webtiles config: banfile '%s' does not exist", path)
    # otherwise, will be a noop on a non-existent file
    l = bans.load_bans(path)
    if l:
        add_ban_list(l)
        early_log("Loading ban list from: '%s'" % path)


# build config.games, as well as various cached bits of game data
def load_game_data(reloading=False):
    global games, game_modes, binaries_to_games
    if not load_games.load_games(reloading=reloading):
        return # failed config reload

    # hopefully we've found some games by this point; do a bit more validation:
    if len(games) == 0:
        raise ValueError("No games defined!")
    if not check_game_config():
        raise ValueError("Errors in game data!")

    # finally, collect info about the games we found (which will also validate
    # the binaries, for versions that support this call at least).
    game_modes = load_games.collect_game_modes()
    binaries_to_games = dict()
    for g in games:
        k = games[g].get_binary_key()
        if k not in binaries_to_games:
            binaries_to_games[k] = []
        binaries_to_games[k].append(g)


def validate():
    check_keys_any(['bind_nonsecure', 'ssl_options'], True)
    if has_key('bind_nonsecure') and get('bind_nonsecure'):
        check_keys_any(['bind_pairs', ['bind_address', 'bind_port']], True)
    if has_key('ssl_options') and get('ssl_options'):
        check_keys_any(['ssl_bind_pairs', ['ssl_address', 'ssl_port']], True)
    if has_key('bind_nonsecure') and get('bind_nonsecure') == "redirect":
        if not check_keys_any(['ssl_options']):
            raise ValueError("bind_nonsecure='redirect' requires ssl ports to redirect to")

    required = ['static_path', 'template_path', 'server_id',
        'dgl_status_file', 'init_player_program',]
    if get('allow_password_reset') or get('admin_password_reset'):
        required.append('lobby_url')

    check_keys_all(required, raise_on_missing=True)

    smpt_opts = ['smtp_host', 'smtp_port', 'smtp_from_addr']
    if check_keys_any(smpt_opts):
        check_keys_all(smpt_opts, True)
    if has_key('smtp_user'):
        check_keys_all('smtp_password', True)

    bans.validate(get('banned'))

    try:
        # simplest is just to run the ban check against a name that isn't
        # valid and so shouldn't ever be banned, which exercises the whole
        # thing
        bans.do_ban_check('', get('banned'))
    except:
        raise ValueError("Webtiles config: malformed ban list ('%s')" %
                                                        repr(get('banned')))

    # set up defaults that are conditioned on other values
    if not has_key('settings_db'):
        set('settings_db', os.path.join(os.path.dirname(get('password_db')),
                                        "user_settings.db3"))



#!/usr/bin/env python3
"""Utility functions for python scripts converting yaml to C++ headers."""

# from __future__ import print_function

# import argparse
import os
import sys
import traceback
import collections
import re

# import yaml  # pip install pyyaml

def quote_or_nullptr(key, d):
    if key in d:
        return quote(d[key])
    else:
        return 'nullptr'


def validate_string(val, name, pattern):
    '''
    Validate a string.

    Note that re.match anchors to the start of the string, so you don't need to
    prefix the pattern with '^'. But it doesn't require matching to the end, so
    you'll probably want to suffix '$'.
    '''
    if not isinstance(val, str):
        raise ValueError('%s isn\'t a string' % name)
    if re.match(pattern, val):
        return val
    else:
        raise ValueError('%s doesn\'t match pattern %s' % (val, pattern))
    return val


def validate_bool(val, name):
    '''Validate a boolean.'''
    if not isinstance(val, bool):
        raise ValueError('%s isn\'t a boolean' % name)
    return val


def validate_int_range(val, name, min, max):
    '''Validate an integer and check it is within the interval [min, max].'''
    if not isinstance(val, int):
        raise ValueError('%s isn\'t an integer' % name)
    if not min <= val <= max:
        raise ValueError('%s isn\'t between %s and %s' % (name, min, max))
    return val


def enumify(name, enum):
    return "{enum}_{name}".format(enum=enum, name=name).replace(' ', '_').upper()


def quote(s):
    if not isinstance(s, str):
        raise ValueError('Expected a string but got %s' % repr(s))
    return '"%s"' % s


def empty_set(typ):
    return "    set<%s>()" % typ


global LIST_TEMPLATE
LIST_TEMPLATE = """    {{ {list} }}"""
global SPELL_LIST_TEMPLATE
SPELL_LIST_TEMPLATE = """    {{\n{list}\n    }}"""

def make_list(list_str, is_spell_list = False):
    global LIST_TEMPLATE
    #TODO: add linebreaks + indents to obey 80 chars?
    if len(list_str.strip()) == 0:
        return "    {}"
    elif is_spell_list:
        return SPELL_LIST_TEMPLATE.format(list=list_str)
    else:
        return LIST_TEMPLATE.format(list=list_str)


def load_template(templatedir, name):
    return open(os.path.join(templatedir, name)).read()



import tornado.web
from tornado.escape import json_encode

import webtiles
from webtiles import config, ws_handler, server

class LobbyHandler(tornado.web.RequestHandler):
    def get(self):
        games = []
        url_base = config.get('lobby_url')
        for s in list(ws_handler.sockets):
            if not s.username or not s.process or not s.show_in_lobby():
                continue

            # XX the following doesn't perfectly match the regular lobby
            # json, arguably one or the other should be changed before this
            # gets to be widely used
            game = dict(
                name=s.username,
                game_id=s.game_id,
                idle_time=s.process.idle_time(),
                viewers=s.process.watcher_count(),
                )
            if url_base:
                # otherwise constructed on js side, see client.js
                game['watch_url'] = url_base + "#watch-" + s.username
            # copy some selected info from `where`. This is a bit more than
            # CrawlProcessHandlerBase.interesting_info.
            # XX game type isn't in here except via game_id, which isn't
            # standardized across servers
            for k in ('v', 'vlong', 'tiles', 'race', 'cls', 'char', 'xl',
                        'title', 'place', 'god', 'turn', 'dur'):
                # the following will force keys not present in the milestone
                # to appear here, e.g. the crawl process doesn't set `god` if
                # there is none. (Maybe this is a bug?)
                game[k] = s.process.where.get(k, "")

            if s.process.last_milestone:
                game["milestone"] = s.process.last_milestone.get("milestone", "")
            else:
                game["milestone"] = ""

            games.append(game)
        self.write(json_encode(games))

class VersionHandler(tornado.web.RequestHandler):
    def get(self):
        self.write(json_encode(server.version_data()))



#!/usr/bin/env python3

"""
Generate species-data.h, aptitudes.h, species-groups.h, and species-type.h

Works with both Python 2 & 3. If that changes, update how the Makefile calls
this.
"""

from __future__ import print_function

import argparse
import os
import sys
import traceback
import re
import collections
if sys.version_info.major == 2:
    from collections import MutableMapping
else:
    from collections.abc import MutableMapping

import yaml  # pip install pyyaml

class Field:
    def __init__(self, f, required = False):
        self.f = f
        self.required = required

ENERGIES = ['walk', 'swim', 'attack', 'missile', 'spell']

class Monster(MutableMapping):
    """Parser for YAML definition files.

    If any YAML content is invalid, the relevant parser function below should
    raise ValueError.
    """

    def __init__(self, yaml_dict):
        self.backing_dict = dict()
        self.from_yaml(yaml_dict)

    def __getitem__(self, key):
        return self.backing_dict[key]

    def __setitem__(self, key, val):
        self.backing_dict[key] = val

    def __delitem__(self, key):
        del self.backing_dict[key]

    def __iter__(self):
        return iter(self.backing_dict)

    def __len__(self):
        return len(self.store)

    def desc_energy(self):
        if 'energy' not in self:
            return 'DEFAULT_ENERGY'
        custom_vals = self['energy']
        return "{%s}" % ', '.join(str(custom_vals.get(e, 10)) for e in ENERGIES)

    def unique_corpse(self): # sorry
        return self["has_corpse"] == 'true' and self["species"] == self["enum"] and 'M_CANT_SPAWN' not in self["flags"]

    def base_tile(self):
        if "tile" not in self:
            return "tilep_" + self["enum"]
        tile = self["tile"]
        if tile.startswith("tile"):
            return tile
        return "tilep_mons_" + tile

    def corpse_tile(self):
        if "corpse_tile" in self:
            return "TILE_CORPSE_" + self["corpse_tile"].upper()
        if not self.unique_corpse():
            return "TILE_ERROR"
        if "tile" in self and self["tile"] != "program_bug":
            return "TILE_CORPSE_" + self["tile"].upper()
        return "TILE_CORPSE_" + self["enum"][len("MONS_"):]

    def fill_tile_desc(self):
        desc = self.base_tile().upper()
        if "tile_variance" in self:
            desc += ", " + self["tile_variance"]
        self["tile_desc"] = desc
        self["tile_corpse"] = self.corpse_tile()

    def from_yaml(self, s):
        for key in s:
            if key not in keyfns:
                raise ValueError("unknown key '%s'" % key)
            if key in self:
                raise ValueError("duplicate key '%s'" % key)
            try:
                self[key] = keyfns[key].f(s[key])
            except ValueError as e:
                raise ValueError("failed to parse '%s': %s" % (key, e))

        for key in keyfns:
            if key not in self and keyfns[key].required:
                raise ValueError("missing required key '%s'" % key)

        for key in defaults:
            if key not in self:
                self[key] = defaults[key]

        if 'enum' not in self:
            self['enum'] = "MONS_" + self['name'].upper().replace(" ", "_")
        if 'species' not in self:
            self['species'] = self['enum']
        if 'genus' not in self:
            self['genus'] = self['species']

        if 'will' not in self:
            if 'will_per_hd' not in self:
                raise ValueError("set exactly one of 'will' and 'will_per_hd'")
            # Yuck, we still use negative will for this?
            self['will'] = -self['will_per_hd']
        elif 'will_per_hd' in self:
            raise ValueError("set exactly one of 'will' and 'will_per_hd'")

        self["joined_flags"] = ' | '.join(self["flags"])
        # Could try to concatenate together resists at the same level, but why?
        self["joined_resists"] = ' | '.join(r.format() for r in self["resists"])
        self["joined_holiness"] = ' | '.join(h for h in self["holiness"])
        self["joined_attacks"] = ', '.join(a.format() for a in self["attacks"])
        self["corpse"] = "true" if "corpse" in self and self["corpse"] else "false" # ew
        self["energy_desc"] = self.desc_energy()
        self.fill_tile_desc()

def get_fields(d, names):
    fields = []
    for f in names:
        if f not in d:
            raise ValueError("missing field '%s'", f)
        fields.append(d[f])
    if len(d) != len(names):
        raise ValueError("unexpected fields") # XXX add more context
    return fields

def parse_bool(s):
    if not isinstance(s, bool):
        raise ValueError("isn't a bool")
    return 'true' if s else 'false'

def parse_str(s):
    if not isinstance(s, str):
        raise ValueError("isn't a string")
    return s

def parse_num(s, min, max):
    if not isinstance(s, int):
        raise ValueError("isn't an int")
    if not min <= s <= max:
        raise ValueError("isn't between %s and %s" % (min, max))
    return s

class Glyph:
    def __init__(self, char, colour):
        self.char = char;
        self.colour = colour;

def parse_glyph(s):
    char, colour = get_fields(s, ["char", "colour"])
    return Glyph(parse_glyph_char(s["char"]), parse_str(s["colour"]).upper())

def parse_glyph_char(s):
    if not isinstance(s, str):
        raise ValueError("glyph isn't a string")
    if len(s) != 1:
        raise ValueError("glyph must be exactly one character")
    return s

def parse_flags(s):
    flags = []
    for f in s:
        flags.append("M_" + f.upper())
    return flags

def parse_resists(s):
    resists = []
    for r in s:
        resists.append(parse_resist(r, s[r]))
    return resists

class Resist:
    def __init__(self, min = 0, max = 1):
        self.min = min;
        self.max = max;

ALL_RESISTS = {
    'fire': Resist(-1, 4),
    'cold': Resist(-1, 4),
    'elec': Resist(-1, 4),
    'poison': Resist( -1, 4),
    'neg': Resist(0, 4),
    'corr': Resist(0, 4),
    'miasma': Resist(0, 4),
    'torment': Resist(),
    'petrify': Resist(),
    'damnation': Resist(),
    'steam': Resist(),
}

class ResVal:
    def __init__(self, enum, val):
        self.enum = enum
        self.val = val

    def format(self):
        if self.val == 0 or self.val == 1:
            return self.enum
        return "mrd(%s, %d)" % (self.enum, self.val)

def parse_resist(name, value):
    if not name in ALL_RESISTS:
        raise ValueError("unknown resist '%s', name")

    res = ALL_RESISTS[name]
    try:
        val = parse_num(value, res.min, res.max)
    except ValueError as e:
        raise ValueError("invalid value for resist '%s': %s" % (name, e))
    return ResVal("MR_RES_%s" % name.upper(), val)

def parse_will(s):
    if s == 'invuln':
        return "WILL_INVULN"
    return parse_num(s, 0, 200)

class Attack:
    def __init__(self, type, damage, flavour):
        self.type = type
        self.damage = damage
        self.flavour = flavour

    def format(self):
        return "{%s, %s, %d}" % (self.type, self.flavour, self.damage)

def parse_attacks(s):
    atks = []
    if not s:
        return atks

    for a in s:
        if 'type' not in a:
            raise ValueError("missing type for attack '%s'", a)
        if 'damage' not in a:
            raise ValueError("missing damage for attack '%s'", a)
        for field in a:
            if field not in {'type', 'damage', 'flavour'}:
                raise ValueError("unknown attack field '%s'", field)
        atks.append(Attack(
            type = "AT_" + a['type'].upper(),
            damage = parse_num(a['damage'], 0, 100),
            flavour = "AF_" + a['flavour'].upper() if 'flavour' in a else 'AF_PLAIN',
        ))
    return atks

def parse_enum(s, prefix, vals):
    if s not in vals:
        raise ValueError("%s not in %s" % (s, vals))
    return prefix + s.upper()

INTELLIGENCES = {'brainless', 'animal', 'human'}
def parse_intelligence(s):
    return parse_enum(s, "I_", INTELLIGENCES)

USE_TYPES = {'nothing', 'open_doors', 'starting_equipment', 'weapons_armour'}
def parse_use(s):
    return parse_enum(s, "MONUSE_", USE_TYPES)

SIZES = {'tiny', 'little', 'small', 'medium', 'large', 'giant'}
def parse_size(s):
    return parse_enum(s, "SIZE_", SIZES)

SHAPES = {
    'humanoid',
    'humanoid_winged',
    'humanoid_tailed',
    'humanoid_winged_tailed',
    'centaur',
    'naga',
    'quadruped',
    'quadruped_tailless',
    'quadruped_winged',
    'bat',
    'bird',
    'snake',
    'fish',
    'insect',
    'insect_winged',
    'arachnid',
    'centipede',
    'snail',
    'plant',
    'fungus',
    'orb',
    'blob',
    'misc',
}
def parse_shape(s):
    return parse_enum(s, "MON_SHAPE_", SHAPES)

def parse_energy(s):
    if 'move' in s:
        if 'walk' in s or 'swim' in s:
            raise ValueError("'move' is redundant with 'walk'/'swim'")
        s['walk'] = s['move']
        s['swim'] = s['move']
        del s['move']
    for k in s:
        if k not in ENERGIES:
            raise ValueError("unknown energy type '%s'", k)
        parse_num(s[k], 1, 100)
    return s

HOLINESSES = {'holy', 'natural', 'undead', 'demonic', 'nonliving', 'plant'}
def parse_holiness(s):
    return [parse_enum(h, "MH_", HOLINESSES) for h in s]

HABITATS = {'land', 'amphibious', 'water', 'lava', 'amphibious_lava',
            'deep_water', 'eldritch_tentacle', 'wall', 'walls_only'}
def parse_habitat(s):
    return parse_enum(s, "HT_", HABITATS)

TVARIES = {'mod', 'cycle', 'random', 'water'}
def parse_tile_variants(s):
    return parse_enum(s, "TVARY_", TVARIES)

keyfns = {
    'glyph': Field(parse_glyph, required = True),
    'name': Field(parse_str, required = True),
    'enum': Field(lambda s: "MONS_" + s.upper()),

    'flags': Field(parse_flags),

    'resists': Field(parse_resists),

    'exp': Field(lambda s: parse_num(s, 0, 9999)),
    'exp_is_mult': Field(parse_bool),
    'species': Field(lambda s: "MONS_" + s.upper()),
    'genus': Field(lambda s: "MONS_" + s.upper()),
    'holiness': Field(parse_holiness),
    'will': Field(parse_will),
    'will_per_hd': Field(lambda s: parse_num(s, 1, 27)),

    'attacks': Field(parse_attacks, required = True),

    'hd': Field(lambda s: parse_num(s, 0, 9999), required = True),
    'hp_10x': Field(lambda s: parse_num(s, 0, 100000000), required = True),

    'ac': Field(lambda s: parse_num(s, -99, 127), required = True),
    'ev': Field(lambda s: parse_num(s, -99, 127), required = True),
    'spells': Field(lambda s: "MST_" + s.upper()),
    'has_corpse': Field(parse_bool),
    'shout': Field(lambda s: "S_" + s.upper()),

    'intelligence': Field(parse_intelligence, required = True),
    'habitat': Field(parse_habitat),
    'speed': Field(lambda s: parse_num(s, 0, 81)),
    'energy': Field(parse_energy),

    'uses': Field(parse_use),
    'size': Field(parse_size, required = True),
    'shape': Field(parse_shape, required = True),
    'god': Field(lambda s: "GOD_" + s.upper().replace(" ", "_")),

    'tile': Field(parse_str),
    'tile_variance': Field(parse_tile_variants),
    'corpse_tile': Field(parse_str),
}

defaults = {
    'flags': ['M_NO_FLAGS'],
    'resists': [ResVal('MR_NO_FLAGS', 0)],
    'exp': 10,
    'exp_is_mult': 'false',
    'holiness': ['MH_NATURAL'],
    'spells': 'MST_NO_SPELLS',
    'shout': 'S_SILENT',
    'habitat': 'HT_LAND',
    'speed': 10,
    'uses': 'MONUSE_NOTHING',
    'has_corpse': 'false',
    'god': 'GOD_NO_GOD',
}

def load_template(templatedir, name):
    return open(os.path.join(templatedir, name)).read()

def main():
    parser = argparse.ArgumentParser(description='Generate mon-data.h')
    parser.add_argument('datadir', help='dat/mons source dir')
    parser.add_argument('templatedir', help='util/mon-gen template source dir')
    parser.add_argument('mon_data', help='mon-data.h output file path')
    args = parser.parse_args()

    if not os.path.isdir(args.datadir):
        print('datadir isn\'t a directory')
        sys.exit(1)
    if not os.path.isdir(args.templatedir):
        print('templatedir isn\'t a directory')
        sys.exit(1)

    text = load_template(args.templatedir, 'header.txt')

    tmpl = load_template(args.templatedir, 'body.txt')
    for f_name in sorted(os.listdir(args.datadir)):
        if not f_name.endswith('.yaml'):
            continue
        f_path = os.path.join(args.datadir, f_name)
        try:
            mon_spec = yaml.safe_load(open(f_path))
        except yaml.YAMLError as e:
            print("Failed to load %s: %s" % (f_name, e))
            sys.exit(1)
        try:
            mons = Monster(mon_spec)
        except (ValueError, KeyError) as e:
            print("Failed to load %s" % f_name)
            traceback.print_exc()
            sys.exit(1)

        text += tmpl.format(**mons)

    text += load_template(args.templatedir, 'footer.txt')

    with open(args.mon_data, 'w') as f:
        f.write(text)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
from __future__ import absolute_import

import os, sys

import webtiles, webtiles.server, webtiles.config

def server_main():
    if sys.version_info.major < 3:
        sys.exit("Python 2 is not supported, please upgrade!")
    # load config values from the traditional module location, which should
    # be in the same directory as this file
    import config as server_config
    webtiles.config.init_config_from_module(server_config)

    # server_path tells where to look for games.d. We can't default this,
    # because this directory is outside the module, and though it is set in
    # the current default `config.py` file, we still need to set it here for
    # backwards-compatibility reasons with dgamelaunch-config setups.
    # TODO: is there a better way to handle this?
    webtiles.config.server_path = os.path.dirname(os.path.abspath(__file__))

    webtiles.server.run()

if __name__ == "__main__":
    server_main()



import itertools
import os, os.path
import yaml

# code for disallowing usernames.


# not an exhaustive leet compendium, but tuned in particular to some common
# slurs. It won't get them all. If someone starts iterating on this, I
# recommend custom code in config.py. Note that punctuation is not generally
# allowed, which removes some headaches here.
leet = {'1': 'i',
        '2': 'r', # also z...
        '3': 'e',
        '4': 'a',
        '5': 's',
        '6': 'g',
        '7': 't',
        '8': 'b',
        '9': 'g',
        '0': 'o',
        }

# return a canonical version of a nickname, doing some deleeting. Useful for
# nickname checking against extreme slurs.
def deleet(s):
    global leet
    # is there a more elegant re version of this?
    c = s.lower()
    for k in leet:
        c = c.replace(k, leet[k])
    return c


# not currently used, but this could be added to as an override list to avoid
# scunthorpe problems
nick_check_override = set()

def nick_check(s, bad_name, leet=False, repeats=False, part=False):
    global nick_check_override
    if bad_name in nick_check_override:
        return False

    if leet:
        s = deleet(s)
    if repeats:
        # remove all repeats from both strings
        s = ''.join(l[0] for l in itertools.groupby(s))
        bad_name = ''.join(l[0] for l in itertools.groupby(bad_name))
    if part:
        # recommended with care: lots of names that you'd want to ban match
        # against parts of many words. (aka
        # https://en.wikipedia.org/wiki/Scunthorpe_problem).
        # But, certain ones such as the n-word are basically safe to do this
        # with, imo. To some extent this issue applies to the other options
        # to this function as well.
        return bad_name.lower() in s.lower()
    else:
        return bad_name.lower() == s.lower()


def do_ban_check(nick, ban_list):
    if not ban_list:
        return False
    for ban_list_element in ban_list:
        # expects ban_list_element to be either:
        # * A str, in which case, we simply check against that nick, or
        # * A dict, containing a mapping 'nicks' to a list of nicks, and
        # optionally an `options` dict that uses `util.nick_check` named
        # arguments to determine how to do the check.
        if not ban_list_element:
            continue
        if isinstance(ban_list_element, str):
            if nick_check(nick, ban_list_element):
                return True
            continue
        ban_list = ban_list_element.get('names', [])
        if not ban_list:
            continue
        for bad in ban_list:
            if nick_check(nick, bad, *ban_list_element.get('options', {})):
                return True
    return False


def validate(ban_list):
    try:
        iter(ban_list)
    except TypeError:
        raise ValueError("Malformed ban list: expected iterable (got '%s')"
                                                        % repr(ban_list))

    try:
        # simplest is just to run the ban check against a name that isn't
        # valid and so shouldn't ever be banned, which exercises the whole
        # thing. In fact, we also throw an error if '' is banned.
        do_ban_check('', ban_list)
    except:
        # try to get a more specific error
        for b in ban_list:
            match = False
            try:
                match = do_ban_check('', [b])
            except:
                raise ValueError("Malformed ban list (element '%s')" % repr(b))
            if match:
                raise ValueError("Ban list matches the empty string!")


# Load a ban list in txt format. This is a list of nicknames, each on their
# own line, that will be matched exactly (non-case-sensitive).
def load_from_txt(path):
    # caller should check path if they are expecting it to definitely exist
    collect = []
    if not os.path.isfile(path):
        return collect
    with open(path) as f:
        for l in f.readlines():
            l = l.strip()
            # allow comments on their own lines only (per sequell), skip blank
            # lines.
            if not l or l[0] == '#':
                continue
            collect.append(l)
    validate(collect)
    return collect


# load a ban list in yaml format. This must be a dict with a single mapping,
# 'banned', to a list. That list consists of either strings (names to ban
# with exact non-case-sensitive matching), or dicts. The dict format is:
#   names: a list of names to check using options:
#   options: can set any of `leet`, `repeats`, or `part to do inexact matching.
#            see `nick_check` for details. As noted there, be wary of
#            overmatching.
# All name matching is non-case-sensitive.
def load_from_yaml(path):
    # caller should check path if they are expecting it to definitely exist
    collect = []
    if not os.path.isfile(path):
        return collect
    with open(path) as f:
        override_data = yaml.safe_load(f)
        # we are expecting a map with a single key, `banned`
        if not isinstance(override_data, dict):
            raise ValueError("Banfile error: '%s' must be a map" % path)
        for key, value in override_data.items():
            # XX could add an 'allowed' list here?
            if key == 'banned':
                if not isinstance(value, list):
                    raise ValueError("Banfile error: '%s' must be a list" % key)
                collect.extend(value)
            else:
                raise ValueError("Banfile error: extraneous key ('%s')" % key)
    validate(collect)
    return collect


def load_bans(path):
    if path.lower().endswith('yml') or path.lower().endswith('yaml'):
        return load_from_yaml(path)
    else:
        return load_from_txt(path)
